import logging
import json
import importlib
import binascii
from typing import List
from types import SimpleNamespace
from neo3.core import cryptography
from .singleton import _Singleton

version = '0.9.3'

core_logger = logging.getLogger('neo3.core')
network_logger = logging.getLogger('neo3.network')
storage_logger = logging.getLogger('neo3.storage')


def load_class_from_path(path_and_class: str):
    """
    Dynamically load a class from a module at the specified path
    Args:
        path_and_class: relative path where to find the module and its class name
        i.e. 'neo3.<package>.<package>.<module>.<class name>'
    Raises:
        ValueError: if the Module or Class is not found.
    Returns:
        class object
    """
    try:
        module_path = '.'.join(path_and_class.split('.')[:-1])
        module = importlib.import_module(module_path)
    except ImportError as err:
        raise ValueError(f"Failed to import module {module_path} with error: {err}")

    try:
        class_name = path_and_class.split('.')[-1]
        class_obj = getattr(module, class_name)
        return class_obj
    except AttributeError as err:
        raise ValueError(f"Failed to get class {class_name} with error: {err}")


class IndexableNamespace(SimpleNamespace):
    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return self.__dict__[key]

    def get(self, key, default=None):
        try:
            return self.__dict__[key]
        except KeyError:
            return default


class Settings(IndexableNamespace):
    db = None
    _cached_standby_committee = None
    default_settings = {
        'network': {
            'magic': 5195086,
            'account_version': 53,
            'seedlist': [],
            'validators_count': 1,
            'standby_committee': ['02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765']
        },
        'storage': {
            'use_default': True,
            'default_provider': 'memory',
            "providers": {
                "memory": {
                    'class_path': 'neo3.storage.implementations.MemoryDB',
                    'options': {}
                },
                "leveldb": {
                    'class_path': 'neo3.storage.implementations.LevelDB',
                    'options': {
                        'path': '/tmp/neo3/'
                    }
                },
                "postgresql": {
                    'class_path': 'neo3.storage.implementations.PostgresDB',
                    'options': {
                        'host': '127.0.0.1',
                        'port': 5432
                    }
                }
            }
        },
        'policy': {
            'max_tx_per_block': 512
        },
        'native_contract_activation': {}
    }

    @classmethod
    def from_json(cls, json: dict):
        o = cls(**json)
        o._convert(o.__dict__, o.__dict__)
        return o

    @classmethod
    def from_file(cls, path_to_json: str):
        with open(path_to_json, 'r') as f:
            data = json.load(f)
        return cls.from_json(data)

    def register(self, json: dict):
        self.__dict__.update(json)
        self._convert(self.__dict__, self.__dict__)

    def _convert(self, what: dict, where: dict):
        # turn all _dictionary what into IndexableNamespaces
        to_update = []
        for k, v in what.items():
            if isinstance(v, dict):
                to_update.append((k, IndexableNamespace(**v)))

        for k, v in to_update:
            if isinstance(where, dict):
                where.update({k: v})
            else:
                where.__dict__.update({k: v})
            self._convert(where[k].__dict__, where[k].__dict__)

    @property
    def standby_committee(self) -> List[cryptography.ECPoint]:
        if self._cached_standby_committee is None:
            points = []
            for p in self.network.standby_committee:
                points.append(cryptography.ECPoint.deserialize_from_bytes(binascii.unhexlify(p)))
            self._cached_standby_committee = points
        return self._cached_standby_committee

    @property
    def standby_validators(self) -> List[cryptography.ECPoint]:
        return self.standby_committee[:self.network.validators_count]

    @property
    def database(self):
        try:
            if self.db:
                return self.db

            if not self.storage.use_default:
                return None

            default_provider_name = self.storage.default_provider
            provider = self.storage.providers[default_provider_name]
            db_class = load_class_from_path(provider.class_path)
            return db_class(provider.options.__dict__)
        except Exception as e:
            return None

    def reset_settings_to_default(self):
        self.__dict__.clear()
        self.__dict__.update(self.from_json(self.default_settings).__dict__)


settings = Settings.from_json(Settings.default_settings)



from __future__ import annotations
import base58  # type: ignore
import base64
import hashlib
import unicodedata
from typing import Optional, Dict, Any, List
from Crypto.Cipher import AES  # type: ignore
from jsonschema import validate  # type: ignore
from neo3 import settings, contracts, vm, wallet, storage
from neo3.network import payloads
from neo3.core import types, to_script_hash, cryptography, syscall_name_to_int


# both constants below are used to encrypt/decrypt a private key to/from a nep2 key
NEP_HEADER = bytes([0x01, 0x42])
NEP_FLAG = bytes([0xe0])
# both constants are used when trying to decrypt a private key from a wif
WIF_PREFIX = bytes([0x80])
WIF_SUFFIX = bytes([0x01])
PRIVATE_KEY_LENGTH = 32


class AccountContract(contracts.Contract):
    _contract_params_schema = {
        "type": ["object", "null"],
        "properties": {
            "name": {"type": "string"},
            "type": {"type": "string"}
        },
        "required": ["name", "type"]
    }
    _json_schema = {
        "type": ["object", "null"],
        "properties": {
            "script": {"type": "string"},
            "parameters": {
                "type": "array",
                "items": _contract_params_schema,
                "minItems": 0,
            },
            "deployed": {"type": "boolean"}
        },
        "required": ["script", "parameters", "deployed"]
    }

    def __init__(self, script: bytes, parameter_list: List[contracts.ContractParameterDefinition]):
        super().__init__(script, [param.type for param in parameter_list])

        self.parameter_names: List[str] = [param.name for param in parameter_list]
        self.deployed: bool = False

    @classmethod
    def from_contract(cls, contract: contracts.Contract) -> AccountContract:
        if isinstance(contract, AccountContract):
            return contract

        parameters = [contracts.ContractParameterDefinition(f"arg{index}", contract.parameter_list[index])
                      for index in range(len(contract.parameter_list))]
        return cls(script=contract.script,
                   parameter_list=parameters)

    @classmethod
    def from_json(cls, json: dict) -> AccountContract:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
        """
        validate(json, schema=cls._json_schema)

        contract = cls(
            script=base64.b64decode(json['script']),
            parameter_list=list(map(lambda p: contracts.ContractParameterDefinition.from_json(p), json['parameters']))
        )
        contract.deployed = json['deployed']

        return contract

    def to_json(self) -> dict:
        """ Convert object into JSON representation. """
        return {
            'script': base64.b64encode(self.script).decode('utf-8'),
            'parameters': list(map(lambda index: {'name': self.parameter_names[index],
                                                  'type': self.parameter_list[index].PascalCase()
                                                  },
                                   range(len(self.parameter_list)))),
            'deployed': self.deployed
        }


class Account:
    _json_schema = {
        "type": "object",
        "properties": {
            "address": {"type": "string"},
            "label": {"type": ["string", "null"]},
            "isDefault": {"type": "boolean"},
            "lock": {"type": "boolean"},
            "key": {"type": ["string", "null"]},
            "contract": AccountContract._json_schema,
            "extra": {"type": ["object", "null"],
                      "properties": {},
                      "additionalProperties": True
                      }
        },
        "required": ["address", "label", "isDefault", "lock", "key", "extra"]
    }

    def __init__(self, password: str,
                 private_key: Optional[bytes] = None,
                 watch_only: bool = False,
                 address: Optional[str] = None,
                 label: Optional[str] = None,
                 lock: bool = False,
                 contract: Optional[contracts.Contract] = None,
                 extra: Optional[Dict[str, Any]] = None
                 ):
        """
        Instantiate an account. This constructor should only be directly called when it's desired to create a new
        account using just a password and a randomly generated private key, otherwise use the alternative constructors
        """
        public_key: Optional[cryptography.ECPoint] = None
        encrypted_key: Optional[bytes] = None
        contract_script: Optional[bytes] = None

        if watch_only:
            if address is None:
                raise ValueError("Creating a watch only account requires an address")
            else:
                self.validate_address(address)

        else:
            key_pair: cryptography.KeyPair

            if private_key is None:
                key_pair = cryptography.KeyPair.generate()
                private_key = key_pair.private_key
            else:
                key_pair = cryptography.KeyPair(private_key)
            encrypted_key = self.private_key_to_nep2(private_key, password)
            contract_script = contracts.Contract.create_signature_redeemscript(key_pair.public_key)
            script_hash = to_script_hash(contract_script)
            address = address if address else self.script_hash_to_address(script_hash)
            public_key = key_pair.public_key

        self.label: Optional[str] = label
        self.address: str = address
        self.public_key = public_key
        self.encrypted_key = encrypted_key
        self.lock = lock

        if not isinstance(contract, AccountContract):
            if contract is not None:
                contract = AccountContract.from_contract(contract)
            elif contract_script is not None:
                default_parameters_list = [
                    contracts.ContractParameterDefinition(name='signature',
                                                          type=contracts.ContractParameterType.SIGNATURE)
                ]
                contract = AccountContract(contract_script, default_parameters_list)

        self.contract: Optional[AccountContract] = contract
        self.extra = extra if extra else {}

    def __eq__(self, other) -> bool:
        return isinstance(other, Account) and self.address == other.address

    @property
    def script_hash(self) -> types.UInt160:
        return self.address_to_script_hash(self.address)

    @property
    def is_watchonly(self) -> bool:
        if self.encrypted_key is None:
            return True
        else:
            return False

    @property
    def is_multisig(self) -> bool:
        if self.contract is None:
            return False
        return contracts.Contract.is_multisig_contract(self.contract.script)

    @property
    def is_single_sig(self) -> bool:
        if self.contract is None:
            return False
        return contracts.Contract.is_signature_contract(self.contract.script)

    def add_as_sender(self, tx: payloads.Transaction):
        """
        Add the account as sender of the transaction.

        Args:
            tx: the transaction to modify
        """
        tx.signers.insert(0, payloads.Signer(self.script_hash, payloads.WitnessScope.GLOBAL))

    def sign_tx(self, tx: payloads.Transaction, password: str, magic: Optional[int] = None) -> None:
        """
        Helper function that signs the TX, adds the Witness and Sender

        Args:
            tx: transaction to sign
            password: the password to decrypt the private key for signing
            magic: the network magic

        Raises:
            ValueError: if transaction validation fails
        """
        if magic is None:
            magic = settings.network.magic

        self._validate_tx(tx)

        message = magic.to_bytes(4, byteorder="little", signed=False) + tx.hash().to_array()
        signature = self.sign(message, password)

        invocation_script = vm.ScriptBuilder().emit_push(signature).to_array()
        # mypy can't infer that the is_watchonly check ensures public_key has a value
        verification_script = contracts.Contract.create_signature_redeemscript(self.public_key)  # type: ignore
        tx.witnesses.insert(0, payloads.Witness(invocation_script, verification_script))

    def sign_multisig_tx(self,
                         tx: payloads.Transaction,
                         password: str,
                         context: wallet.MultiSigContext,
                         magic: Optional[int] = None) -> None:
        if magic is None:
            magic = settings.network.magic

        if not self.contract:
            raise ValueError("Account is not a valid multi-signature account")

        # When importing a multi-sig account it searches for an associated regular account to copy key material from.
        # However, it is possible to add a multi-sig account before having a regular account with one of the required
        # public keys for the multi-sig account. Therefore, we should check if we actually have key material to continue
        if self.is_watchonly:
            _, _, public_keys = contracts.Contract.parse_as_multisig_contract(self.contract.script)
            raise ValueError(f"Cannot sign with watch only account. Try adding a regular account to your wallet "
                             f"matching one of the following public keys, or update the key material for this account "
                             f"directly."
                             f" {list(map(lambda pk: str(pk), public_keys))}")

        self._validate_tx(tx)

        if not self.is_multisig:
            raise ValueError("Account is not a valid multi-signature account")

        if not context.initialised:
            context.process_contract(self.contract.script)

        if self.public_key not in context.expected_public_keys:
            raise ValueError("Account is not in the required key list for this signing context")

        message = magic.to_bytes(4, byteorder="little", signed=False) + tx.hash().to_array()
        signature = self.sign(message, password)

        context.signature_pairs.update({self.public_key: signature})

        if context.is_complete:
            # build and insert multisig witness
            sb = vm.ScriptBuilder()

            pairs = list(context.signature_pairs.items())
            # sort by public key
            pairs.sort(key=lambda p: p[0])

            for i, (key, sig) in enumerate(pairs):
                if i == context.signing_threshold:
                    break
                sb.emit_push(sig)

            invocation_script = sb.to_array()
            verification_script = contracts.Contract.create_multisig_redeemscript(context.signing_threshold,
                                                                                  context.expected_public_keys)

            tx.witnesses.insert(0, payloads.Witness(invocation_script, verification_script))

    def sign(self, data: bytes, password: str) -> bytes:
        """
        Sign arbitrary data using the SECP256R1 curve.

        Args:
            data: data to be signed
            password: the password to decrypt the private key

        Returns:
            signature of the signed data
        """
        if self.is_watchonly:
            raise ValueError("Cannot sign transaction using a watch only account")
        # mypy can't infer that the is_watchonly check ensures encrypted_key has a value
        private_key = self.private_key_from_nep2(self.encrypted_key.decode("utf-8"), password)  # type: ignore
        return cryptography.sign(data, private_key)

    @classmethod
    def create_new(cls, password: str) -> Account:
        return cls(password=password, watch_only=False)

    @classmethod
    def from_encrypted_key(cls, nep2_key: str, password: str) -> Account:
        """
        Instantiate and returns an account from a given nep2 key and password.

        Args:
            nep2_key: the encrypted private key.
            password: the password to decrypt the nep2 key.

        Returns:
            The newly created account.
        """
        return cls(password=password, private_key=cls.private_key_from_nep2(nep2_key, password))

    @classmethod
    def from_private_key(cls, private_key: bytes, password: str) -> Account:
        """
        Instantiate and returns an account from a given private key and password.

        Args:
            private_key: the private key that will be used to create an encrypted key.
            password: the password to encrypt a randomly generated private key.

        Returns:
            The newly created account.
        """
        return cls(password=password, private_key=private_key)

    @classmethod
    def from_wif(cls, wif: str, password: str) -> Account:
        """
        Instantiate and returns an account from a given wif and password.

        Args:
            wif: the wif that will be decrypted to get a private key and generate an encrypted key.
            password: the password to encrypt the private key with.

        Returns:
            The newly created account.
        """
        return cls(password=password, private_key=cls.private_key_from_wif(wif))

    @classmethod
    def watch_only(cls, script_hash: types.UInt160) -> Account:
        """
        Instantiate and returns a watch-only account from a given script hash.

        Args:
            script_hash: the script hash that will identify an account to be watched.

        Returns:
            The account that will be monitored.
        """
        return cls(password='', watch_only=True, address=cls.script_hash_to_address(script_hash))

    @classmethod
    def watch_only_from_address(cls, address: str) -> Account:
        """
        Instantiate and returns a watch-only account from a given address.

        Args:
            address: the address that will identify an account to be watched.

        Returns:
            The account that will be monitored.
        """
        return cls(password='', watch_only=True, address=address)

    def to_json(self) -> dict:
        return {
            'address': self.address,
            'label': self.label,
            'lock': self.lock,
            'key': self.encrypted_key.decode('utf-8') if self.encrypted_key is not None else None,
            'contract': self.contract.to_json() if self.contract is not None else None,
            'extra': self.extra if len(self.extra) > 0 else None
        }

    @classmethod
    def from_json(cls, json: dict, password: str) -> Account:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
            password: the password to decrypt the json data.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
        """
        validate(json, schema=cls._json_schema)

        return cls(password=password,
                   private_key=(cls.private_key_from_nep2(json['key'], password)
                                if json['key'] is not None else json['key']),
                   address=json['address'],
                   label=json['label'],
                   lock=json['lock'],
                   contract=AccountContract.from_json(json['contract']),
                   extra=json['extra']
                   )

    @staticmethod
    def script_hash_to_address(script_hash: types.UInt160) -> str:
        """
        Converts the specified script hash to an address.

        Args:
            script_hash: script hash to convert.
        """
        version = settings.network.account_version  # this is the current Neo's protocol version
        data = version.to_bytes(1, 'little') + script_hash.to_array()

        return base58.b58encode_check(data).decode('utf-8')

    @staticmethod
    def address_to_script_hash(address: str) -> types.UInt160:
        """
        Converts the specified address to a script hash.

        Args:
            address: address to convert

        Raises:
            ValueError: if the length of data (address value in bytes) is not valid.
            ValueError: if the account version is not valid.
        """
        Account.validate_address(address)

        data = base58.b58decode_check(address)

        return types.UInt160(data[1:])

    @staticmethod
    def private_key_from_nep2(nep2_key: str, passphrase: str,
                              scrypt_parameters: Optional[wallet.ScryptParameters] = None) -> bytes:
        """
        Decrypt a nep2 key into a private key.

        Args:
            nep2_key: the key that will be decrypt.
            passphrase: the password to decrypt the nep2 key.
            scrypt_parameters: a ScryptParameters object that will be passed to the key derivation function.

        Raises:
            ValueError: if the length of the nep2_key is not valid.
            ValueError: if it's not possible to decode the nep2_key.
            ValueError: if the passphrase is incorrect or the version of the account is not valid.

        Returns:
            the private key.
        """
        if scrypt_parameters is None:
            scrypt_parameters = wallet.ScryptParameters()

        if len(nep2_key) != 58:
            raise ValueError(f"Please provide a nep2_key with a length of 58 bytes (LEN: {len(nep2_key)})")

        address_hash_size = 4
        address_hash_offset = len(NEP_FLAG) + len(NEP_HEADER)

        try:
            decoded_key = base58.b58decode_check(nep2_key)
        except Exception:
            raise ValueError("Base58decode failure of nep2 key")

        address_checksum = decoded_key[address_hash_offset:address_hash_offset + address_hash_size]
        encrypted = decoded_key[-32:]

        pwd_normalized = bytes(unicodedata.normalize("NFC", passphrase), "utf-8")
        derived = hashlib.scrypt(password=pwd_normalized, salt=address_checksum,
                                 n=scrypt_parameters.n,
                                 r=scrypt_parameters.r,
                                 p=scrypt_parameters.p,
                                 dklen=64)

        derived1 = derived[:32]
        derived2 = derived[32:]

        cipher = AES.new(derived2, AES.MODE_ECB)
        decrypted = cipher.decrypt(encrypted)
        private_key = Account._xor_bytes(decrypted, derived1)

        # Now check that the address hashes match. If they don't, the password was wrong.
        key_pair = cryptography.KeyPair(private_key=private_key)
        script_hash = to_script_hash(contracts.Contract.create_signature_redeemscript(key_pair.public_key))
        address = Account.script_hash_to_address(script_hash)
        first_hash = hashlib.sha256(address.encode("utf-8")).digest()
        second_hash = hashlib.sha256(first_hash).digest()
        checksum = second_hash[:4]
        if checksum != address_checksum:
            raise ValueError(f"Wrong passphrase or key was encrypted with an address version that is not "
                             f"{settings.network.account_version}")

        return private_key

    @staticmethod
    def private_key_to_nep2(private_key: bytes, passphrase: str,
                            scrypt_parameters: Optional[wallet.ScryptParameters] = None) -> bytes:
        """
        Encrypt a private key into a nep2 key.

        Args:
            private_key: the key that will be encrypt.
            passphrase: the password to encrypt the nep2 key.
            scrypt_parameters: a ScryptParameters object that will be passed to the key derivation function.

        Returns:
            the encrypted nep2 key.
        """
        if scrypt_parameters is None:
            scrypt_parameters = wallet.ScryptParameters()

        key_pair = cryptography.KeyPair(private_key=private_key)
        script_hash = to_script_hash(contracts.Contract.create_signature_redeemscript(key_pair.public_key))
        address = Account.script_hash_to_address(script_hash)
        # NEP2 checksum: hash the address twice and get the first 4 bytes
        first_hash = hashlib.sha256(address.encode("utf-8")).digest()
        second_hash = hashlib.sha256(first_hash).digest()
        checksum = second_hash[:4]

        pwd_normalized = bytes(unicodedata.normalize("NFC", passphrase), "utf-8")
        derived = hashlib.scrypt(password=pwd_normalized, salt=checksum,
                                 n=scrypt_parameters.n,
                                 r=scrypt_parameters.r,
                                 p=scrypt_parameters.p,
                                 dklen=64)

        derived1 = derived[:32]
        derived2 = derived[32:]

        xor_ed = Account._xor_bytes(bytes(private_key), derived1)
        cipher = AES.new(derived2, AES.MODE_ECB)
        encrypted = cipher.encrypt(xor_ed)

        nep2 = bytearray()
        nep2.extend(NEP_HEADER)
        nep2.extend(NEP_FLAG)
        nep2.extend(checksum)
        nep2.extend(encrypted)

        # Finally, encode with Base58Check
        encoded_nep2 = base58.b58encode_check(bytes(nep2))

        return encoded_nep2

    @staticmethod
    def private_key_from_wif(wif: str) -> bytes:
        """
        Decrypt a private key from a wif.

        Args:
            wif: the wif that will be decrypted.

        Raises:
            ValueError: if the wif is not valid.
        """
        try:
            decoded_key: bytes = base58.b58decode_check(wif)
        except Exception:
            raise ValueError("Base58decode failure of wif")

        if len(decoded_key) != 34:
            raise ValueError(f"The decoded wif length should be "
                             f"{len(WIF_PREFIX) + PRIVATE_KEY_LENGTH + len(WIF_SUFFIX)}, while the given wif "
                             f"length is {len(decoded_key)}")
        elif decoded_key[:1] != WIF_PREFIX:
            raise ValueError(f"The decoded wif first byte should be {str(WIF_PREFIX)}")
        elif decoded_key[-1:] != WIF_SUFFIX:
            raise ValueError(f"The decoded wif last byte should be {str(WIF_SUFFIX)}")

        private_key = decoded_key[1: 33]

        return private_key

    @staticmethod
    def is_valid_address(address: str) -> bool:
        """
        Test if the provided address is a valid address.

        Args:
            address: an address.
        """
        try:
            Account.validate_address(address)
        except ValueError:
            return False
        return True

    @staticmethod
    def validate_address(address: str) -> None:
        """
        Validate a given address. If address is not valid an exception will be raised.

        Args:
            address: an address.

        Raises:
            ValueError: if the length of data(address value in bytes) is not valid.
            ValueError: if the account version is not valid.
        """
        data: bytes = base58.b58decode_check(address)
        if len(data) != len(types.UInt160.zero()) + 1:
            raise ValueError(f"The address is wrong, because data (address value in bytes) length should be "
                             f"{len(types.UInt160.zero()) + 1}")
        elif data[0] != settings.network.account_version:
            raise ValueError(f"The account version is not {settings.network.account_version}")

    @staticmethod
    def _xor_bytes(a: bytes, b: bytes) -> bytes:
        """
        XOR on two bytes objects
        Args:
            a (bytes): object 1
            b (bytes): object 2
        Returns:
            bytes: The XOR result
        """
        assert len(a) == len(b)
        res = bytearray()
        for i in range(len(a)):
            res.append(a[i] ^ b[i])
        return bytes(res)

    def _validate_tx(self, tx: payloads.Transaction) -> None:
        """
        Helper to validate properties before signing
        """
        if tx.network_fee == 0 or tx.system_fee == 0:
            raise ValueError("Transaction validation failure - "
                             "a transaction without network and system fees will always fail to validate on chain")

        if len(tx.signers) == 0:
            raise ValueError("Transaction validation failure - Missing sender")

        if len(tx.script) == 0:
            raise ValueError("Transaction validation failure - script field can't be empty")

        if self.is_watchonly:
            raise ValueError("Cannot sign transaction using a watch only account")



from __future__ import annotations
from .service import InteropService, InteropDescriptor
# having the decorator out of the __init__ ensures we can use it in the modules below on import
# otherwise we get a circular import issue
from .decorator import register
# the __name__ imports are just to trigger module loading,
# which in turn executes the decorators to register the SYSCALLS
from .contract import __name__
from .crypto import __name__
from .enumerator import IIterator, StorageIterator
from .runtime import __name__
from .storage import storage_put, MAX_STORAGE_VALUE_SIZE, MAX_STORAGE_KEY_SIZE



import unittest
from neo3 import contracts
from neo3.core import cryptography, types


class ContractPermissionDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        private_key = b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
        cls.keypair = cryptography.KeyPair(private_key)

    def test_wildcard(self):
        cpd = contracts.ContractPermissionDescriptor()
        self.assertTrue(cpd.is_wildcard)
        self.assertDictEqual({'contract': '*'}, cpd.to_json())
        cpd_from_json = contracts.ContractPermissionDescriptor.from_json(cpd.to_json())
        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_group(self):
        cpd = contracts.ContractPermissionDescriptor(contract_hash=types.UInt160.zero())
        self.assertTrue(cpd.is_hash)
        self.assertFalse(cpd.is_group)
        self.assertFalse(cpd.is_wildcard)
        self.assertDictEqual({'contract': '0x0000000000000000000000000000000000000000'}, cpd.to_json())
        cpd_from_json = contracts.ContractPermissionDescriptor.from_json(cpd.to_json())
        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_contract_hash(self):
        cpd = contracts.ContractPermissionDescriptor(group=self.keypair.public_key)
        self.assertFalse(cpd.is_hash)
        self.assertTrue(cpd.is_group)
        self.assertFalse(cpd.is_wildcard)
        self.assertDictEqual({'contract': '033d523f36a732974c0f7dbdfafb5206ecd087211366a274190f05b86d357f4bad'},
                             cpd.to_json())
        cpd_from_json = contracts.ContractPermissionDescriptor.from_json(cpd.to_json())

        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_exceptions(self):
        # test construction with too many arguments given
        with self.assertRaises(ValueError) as context:
            cpd = contracts.ContractPermissionDescriptor(
                contract_hash=types.UInt160.zero(),
                group=self.keypair.public_key)
        self.assertIn("Maximum 1 argument", str(context.exception))

        # test from_json with invalid json
        with self.assertRaises(ValueError) as context:
            contracts.ContractPermissionDescriptor.from_json({'contract':None})
        self.assertEqual("Invalid JSON - Cannot deduce permission type from None", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.ContractPermissionDescriptor.from_json({'contract': 'abc'})
        self.assertEqual("Invalid JSON - Cannot deduce permission type from: abc", str(context.exception))




import unittest
import binascii
import hashlib
import mmh3
from neo3.core import types
from neo3.core import cryptography as crypto


class MerkleTreeTestCase(unittest.TestCase):
    def test_compute_root_single_hash(self):
        data = binascii.unhexlify(b'aa' * 32)
        hash1 = types.UInt256(data=data)
        root = crypto.MerkleTree.compute_root([hash1])

        self.assertEqual(data, root.to_array())

    def test_compute_root_multiple_hashes(self):
        expected_hash = hashlib.sha256(hashlib.sha256(binascii.unhexlify(b'aa' * 32 + b'bb' * 32)).digest()).digest()

        hash1 = types.UInt256(data=binascii.unhexlify(b'aa' * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b'bb' * 32))
        hashes = [hash1, hash2]
        root = crypto.MerkleTree.compute_root(hashes)

        self.assertEqual(expected_hash, root.to_array())

    def test_computer_root_no_input(self):
        self.assertEqual(types.UInt256.zero(), crypto.MerkleTree.compute_root([]))

    def test_build_no_leaves(self):
        with self.assertRaises(ValueError) as context:
            crypto.MerkleTree._build([])
        self.assertIn("Leaves must have length", str(context.exception))

    def test_to_hash_array(self):
        hash1 = types.UInt256(data=binascii.unhexlify(b'aa' * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b'bb' * 32))
        hash3 = types.UInt256(data=binascii.unhexlify(b'cc' * 32))
        hash4 = types.UInt256(data=binascii.unhexlify(b'dd' * 32))
        hash5 = types.UInt256(data=binascii.unhexlify(b'ee' * 32))
        hashes = [hash1, hash2, hash3, hash4, hash5]

        m = crypto.MerkleTree(hashes)
        hash_array = m.to_hash_array()

        # sort the array
        hash_array = sorted(hash_array)

        for i, h in enumerate(hashes):
            self.assertEqual(h, hash_array[i])

    def test_merkle_node_methods(self):
        hash1 = types.UInt256(data=binascii.unhexlify(b'aa' * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b'bb' * 32))
        hashes = [hash1, hash2]
        m = crypto.MerkleTree(hashes)

        self.assertEqual(True, m.root.is_root())
        self.assertEqual(False, m.root.is_leaf())
        self.assertEqual(False, m.root.left_child.is_root())
        self.assertEqual(True, m.root.left_child.is_leaf())

class BloomFilterTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_seed_building(self):
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456)
        # these values have been captured by creating a BloomFilter in C# and reading the seed values through debugging
        expected_seeds = [123456, 4222003669, 4148916586, 4075829503, 4002742420, 3929655337, 3856568254, 3783481171,
                          3710394088, 3637307005]
        self.assertEqual(expected_seeds, filter.seeds)

    def test_add_and_check(self):
        # modelled after https://github.com/neo-project/neo/blob/982e69090f27c1415872536ce39aea22f0873467/neo.UnitTests/Cryptography/UT_BloomFilter.cs#L12
        elements = b'\x00\x01\x02\x03\x04'
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456)
        filter.add(elements)
        self.assertTrue(filter.check(elements))
        self.assertFalse(filter.check(elements+b'\x05\x06\x07\x08\x09'))

    def test_init_with_elements(self):
        elements = b'\x00\x01\x02\x03\x04'
        m=7
        k=10
        filter = crypto.BloomFilter(m=m, k=k, ntweak=123456, elements=elements)
        self.assertEqual(m, len(filter.bits))
        self.assertEqual(k, len(filter.seeds))

    def test_get_bits(self):
        """
        byte[] elements = { 0, 1, 2, 3, 4};
        BloomFilter bf = new BloomFilter(7, 10, 123456, elements);
        byte[] buffer = new byte[(bf.M/8)+1];
        Console.WriteLine($"\\x{BitConverter.ToString(buffer).Replace("-", "\\x")}");
        """
        elements = b'\x00\x01\x02\x03\x04'
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456, elements=elements)
        self.assertEqual(b'\x00', filter.get_bits())

class Murmur128test(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_neo_cases(self):
        # https://github.com/Liaojinghui/neo/blob/30f33d075502acd792f804ffcf84cce689255306/tests/neo.UnitTests/Cryptography/UT_Murmur128.cs
        self.assertEqual(bytes.fromhex("0bc59d0ad25fde2982ed65af61227a0e"), mmh3.hash_bytes("hello", 123))
        self.assertEqual(bytes.fromhex("3d3810fed480472bd214a14023bb407f"), mmh3.hash_bytes("world", 123))
        self.assertEqual(bytes.fromhex("e0a0632d4f51302c55e3b3e48d28795d"), mmh3.hash_bytes("hello world", 123))


from __future__ import annotations
from typing import Optional, cast, List
from . import NativeContract, register
from neo3 import contracts, storage, vm
from neo3.core import types, serialization, to_script_hash, msgrouter
from neo3.network import payloads


class OracleRequest(serialization.ISerializable):
    def __init__(self,
                 original_tx_id: types.UInt256,
                 gas_for_response: int,
                 url: str,
                 filter: str,
                 callback_contract: types.UInt160,
                 callback_method: str,
                 user_data: bytes
                 ):
        self.original_tx_id = original_tx_id
        self.gas_for_response = gas_for_response
        self.url = url
        self.filter = filter
        self.callback_contract = callback_contract
        self.callback_method = callback_method
        self.user_data = user_data

    def __len__(self):
        return len(self.to_array())

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable(self.original_tx_id)
        writer.write_uint64(self.gas_for_response)
        writer.write_var_string(self.url)
        writer.write_var_string(self.filter)
        writer.write_serializable(self.callback_contract)
        writer.write_var_string(self.callback_method)
        writer.write_var_bytes(self.user_data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.original_tx_id = reader.read_serializable(types.UInt256)
        self.gas_for_response = reader.read_uint64()
        self.url = reader.read_var_string()
        self.filter = reader.read_var_string()
        self.callback_contract = reader.read_serializable(types.UInt160)
        self.callback_method = reader.read_var_string()
        self.user_data = reader.read_var_bytes()

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt256.zero(), 0, "", "", types.UInt160.zero(), "", b'')


class OracleContract(NativeContract):
    _MAX_URL_LENGTH = 256
    _MAX_FILTER_LEN = 128
    _MAX_CALLBACK_LEN = 32
    _MAX_USER_DATA_LEN = 512
    _id = -9

    key_request_id = storage.StorageKey(_id, b'\x09')
    key_request = storage.StorageKey(_id, b'\x07')
    key_id_list = storage.StorageKey(_id, b'\x06')
    key_price = storage.StorageKey(_id, b'\x05')

    def init(self):
        super(OracleContract, self).init()
        self.manifest.abi.events = [
            contracts.ContractEventDescriptor(
                "OracleRequest",
                parameters=[
                    contracts.ContractParameterDefinition("Id", contracts.ContractParameterType.INTEGER),
                    contracts.ContractParameterDefinition("RequestContract", contracts.ContractParameterType.HASH160),
                    contracts.ContractParameterDefinition("Url", contracts.ContractParameterType.STRING),
                    contracts.ContractParameterDefinition("Filter", contracts.ContractParameterType.STRING)
                ]
            ),
            contracts.ContractEventDescriptor(
                "OracleResponse",
                parameters=[
                    contracts.ContractParameterDefinition("Id", contracts.ContractParameterType.INTEGER),
                    contracts.ContractParameterDefinition("OriginalTx", contracts.ContractParameterType.HASH160)
                ]
            )
        ]

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        engine.snapshot.storages.put(self.key_request_id, storage.StorageItem(vm.BigInteger.zero().to_array()))
        engine.snapshot.storages.put(self.key_price, storage.StorageItem(vm.BigInteger(50000000).to_array()))

    @register("setPrice", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def set_price(self, engine: contracts.ApplicationEngine, price: int) -> None:
        if price <= 0:
            raise ValueError("Oracle->setPrice value cannot be negative or zero")
        if not self._check_committee(engine):
            raise ValueError("Oracle->setPrice check committee failed")
        item = engine.snapshot.storages.get(self.key_price)
        item.value = vm.BigInteger(price).to_array()

    @register("getPrice", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_price(self, snapshot: storage.Snapshot) -> int:
        return int(vm.BigInteger(snapshot.storages.get(self.key_price, read_only=True).value))

    @register("finish",
              (contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_CALL | contracts.CallFlags.ALLOW_NOTIFY))
    def finish(self, engine: contracts.ApplicationEngine) -> None:
        tx = engine.script_container
        tx = cast(payloads.Transaction, tx)
        response = tx.try_get_attribute(payloads.OracleResponse)
        if response is None:
            raise ValueError("Oracle response not found")

        request = self.get_request(engine.snapshot, response.id)
        if request is None:
            raise ValueError("Oracle request not found")

        state = vm.ArrayStackItem(
            engine.reference_counter,
            [vm.IntegerStackItem(response.id),
             vm.ByteStringStackItem(request.original_tx_id.to_array())
             ]
        )

        msgrouter.interop_notify(self.hash, "OracleResponse", state)

        user_data = contracts.BinarySerializer.deserialize(request.user_data,
                                                           engine.MAX_STACK_SIZE,
                                                           engine.reference_counter)
        args: List[vm.StackItem] = [vm.ByteStringStackItem(request.url.encode()),
                                    user_data,
                                    vm.IntegerStackItem(int(response.code)),
                                    vm.ByteStringStackItem(response.result)]

        engine.call_from_native(self.hash, request.callback_contract, request.callback_method, args)

    @register("request", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY)
    def _request(self,
                 engine: contracts.ApplicationEngine,
                 url: str,
                 filter: str,
                 callback: str,
                 user_data: vm.StackItem,
                 gas_for_response: int) -> None:
        if len(url.encode('utf-8')) > self._MAX_URL_LENGTH or \
                len(filter.encode('utf-8')) > self._MAX_FILTER_LEN or \
                len(callback.encode('utf-8')) > self._MAX_CALLBACK_LEN or \
                callback.startswith("_") or \
                gas_for_response < 10000000:
            raise ValueError

        engine.add_gas(self.get_price(engine.snapshot))
        engine.add_gas(gas_for_response)
        self._gas.mint(engine, self.hash, vm.BigInteger(gas_for_response), False)

        si_item_id = engine.snapshot.storages.get(self.key_request_id, read_only=False)
        item_id = vm.BigInteger(si_item_id.value)
        si_item_id.value = (item_id + 1).to_array()

        if contracts.ManagementContract().get_contract(engine.snapshot, engine.calling_scripthash) is None:
            raise ValueError

        oracle_request = OracleRequest(self._get_original_txid(engine),
                                       gas_for_response,
                                       url,
                                       filter,
                                       engine.calling_scripthash,
                                       callback,
                                       contracts.BinarySerializer.serialize(user_data, self._MAX_USER_DATA_LEN))
        engine.snapshot.storages.put(self.key_request + int(item_id).to_bytes(8, 'little', signed=False),
                                     storage.StorageItem(oracle_request.to_array())
                                     )

        sk_id_list = self.key_id_list + self._get_url_hash(url)
        si_id_list = engine.snapshot.storages.try_get(sk_id_list, read_only=False)
        if si_id_list is None:
            si_id_list = storage.StorageItem(b'\x00')

        with serialization.BinaryReader(si_id_list.value) as reader:
            count = reader.read_var_int()
            id_list = []
            for _ in range(count):
                id_list.append(reader.read_uint64())

        id_list.append(item_id)
        if len(id_list) >= 256:
            raise ValueError("Oracle has too many pending responses for this url")

        with serialization.BinaryWriter() as writer:
            writer.write_var_int(len(id_list))
            for id in id_list:
                writer.write_uint64(id)
            si_id_list.value = writer.to_array()
        engine.snapshot.storages.update(sk_id_list, si_id_list)

        state = vm.ArrayStackItem(
            engine.reference_counter,
            [vm.IntegerStackItem(item_id),
             vm.ByteStringStackItem(engine.calling_scripthash.to_array()),
             vm.ByteStringStackItem(url.encode()),
             vm.ByteStringStackItem(filter.encode()),
             ]
        )

        msgrouter.interop_notify(self.hash, "OracleRequest", state)

    @register("verify", contracts.CallFlags.READ_ONLY, cpu_price=1 << 15)
    def _verify(self, engine: contracts.ApplicationEngine) -> bool:
        tx = engine.script_container
        if not isinstance(tx, payloads.Transaction):
            return False
        return bool(tx.try_get_attribute(payloads.OracleResponse))

    def get_request(self, snapshot: storage.Snapshot, id: int) -> Optional[OracleRequest]:
        id_bytes = id.to_bytes(8, 'little', signed=False)
        storage_item = snapshot.storages.try_get(self.key_request + id_bytes)
        if storage_item is None:
            return None

        return OracleRequest.deserialize_from_bytes(storage_item.value)

    def post_persist(self, engine: contracts.ApplicationEngine) -> None:
        super(OracleContract, self).post_persist(engine)
        nodes = []
        for tx in engine.snapshot.persisting_block.transactions:
            response = tx.try_get_attribute(payloads.OracleResponse)
            if response is None:
                continue

            # remove request from storage
            sk_request = self.key_request + response.id.to_bytes(8, 'little')
            si_request = engine.snapshot.storages.try_get(sk_request)
            if si_request is None:
                continue
            request = OracleRequest.deserialize_from_bytes(si_request.value)
            engine.snapshot.storages.delete(sk_request)

            # remove id from id list
            sk_id_list = self.key_id_list + self._get_url_hash(request.url)
            si_id_list = engine.snapshot.storages.try_get(sk_id_list, read_only=False)
            if si_id_list is None:
                si_id_list = storage.StorageItem(b'\x00')

            try:
                id_list = si_id_list.get(_IdList)
            except Exception as e:
                print(e)
            id_list.remove(response.id)
            if len(id_list) == 0:
                engine.snapshot.storages.delete(sk_id_list)

            # mint gas for oracle nodes
            nodes_public_keys = contracts.DesignationContract().get_designated_by_role(
                engine.snapshot,
                contracts.DesignateRole.ORACLE,
                engine.snapshot.persisting_block.index)

            for public_key in nodes_public_keys:
                nodes.append([
                    to_script_hash(contracts.Contract.create_signature_redeemscript(public_key)),
                    vm.BigInteger.zero()
                ])
            if len(nodes) > 0:
                idx = response.id % len(nodes)
                # mypy can't figure out that the second item is a BigInteger
                nodes[idx][1] += self.get_price(engine.snapshot)  # type: ignore

        for pair in nodes:
            if pair[1].sign > 0:  # type: ignore
                self._gas.mint(engine, pair[0], pair[1], False)

    def _get_url_hash(self, url: str) -> bytes:
        return to_script_hash(url.encode('utf-8')).to_array()

    def _get_original_txid(self, engine: contracts.ApplicationEngine) -> types.UInt256:
        tx = cast(payloads.Transaction, engine.script_container)
        response = tx.try_get_attribute(payloads.OracleResponse)
        if response is None:
            return tx.hash()
        request = self.get_request(engine.snapshot, response.id)
        if request is None:
            raise ValueError  # C# will throw null pointer access exception
        return request.original_tx_id


class _IdList(list, serialization.ISerializable):
    """
    Helper class to get an IdList from storage and deal with caching.
    """
    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_var_int(len(self))
        for item in self:
            writer.write_uint64(item)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        count = reader.read_var_int()
        for _ in range(count):
            self.append(reader.read_uint64())



from __future__ import annotations
from typing import Dict, Optional
from neo3.network import payloads
from neo3.core import types, msgrouter
from neo3 import network_logger as logger, _Singleton
from contextlib import suppress


class RelayCache(_Singleton):
    """
    A cache holding transactions broadcasted to the network to be included in a block.

    Will be accessed in response to a GETDATA network payload.
    """
    def init(self):
        self.cache: Dict[types.UInt256, payloads.inventory.IInventory] = dict()
        msgrouter.on_block_persisted += self.update_cache_for_block_persist

    def add(self, inventory: payloads.inventory.IInventory) -> None:
        self.cache.update({inventory.hash(): inventory})

    def get_and_remove(self, inventory_hash: types.UInt256) -> Optional[payloads.inventory.IInventory]:
        try:
            return self.cache.pop(inventory_hash)
        except KeyError:
            return None

    def try_get(self, inventory_hash: types.UInt256) -> Optional[payloads.inventory.IInventory]:
        return self.cache.get(inventory_hash, None)

    def update_cache_for_block_persist(self, block: payloads.Block) -> None:
        for tx in block.transactions:
            with suppress(KeyError):
                self.cache.pop(tx.hash())
                logger.debug(f"Found {tx.hash()} in last persisted block. Removing from relay cache")

    def reset(self) -> None:
        self.cache = dict()



from __future__ import annotations
import binascii
from neo3.core import IJson, types, cryptography


class ContractPermissionDescriptor(IJson):
    """
    A restriction object that limits the smart contract's calling abilities. Enforced at runtime.

    See Also: ContractManifest.
    """
    def __init__(self, contract_hash: types.UInt160 = None,
                 group: cryptography.ECPoint = None):
        """
        Create a contract hash or group based restriction. Mutually exclusive.
        Supply no arguments to create a wildcard permission descriptor.

        Raises:
            ValueError: if both contract hash and group arguments are supplied.
        """
        if contract_hash is not None and group is not None:
            raise ValueError("Maximum 1 argument can be supplied")
        self.contract_hash = contract_hash
        self.group = group

    def __eq__(self, other):
        return self.contract_hash == other.contract_hash and self.group == other.group

    @property
    def is_hash(self) -> bool:
        """
        Indicates if the permission is limited to a specific contract hash.
        """
        return self.contract_hash is not None

    @property
    def is_group(self) -> bool:
        """
        Indicates if the permission is limited to a specific group.
        """
        return self.group is not None

    @property
    def is_wildcard(self) -> bool:
        """
        Indicates if the permission is not limited to a specific contract or a specific group.
        """
        return not self.is_hash and not self.is_group

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        # NEO C# deviates here. They return a string
        if self.contract_hash:
            val = "0x" + str(self.contract_hash)
        elif self.group:
            val = str(self.group)
        else:
            val = "*"
        return {'contract': val}

    @classmethod
    def from_json(cls, json: dict) -> ContractPermissionDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        # catches both missing key and None as value
        value = json.get('contract', None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot deduce permission type from None")

        if len(value) == 42:
            return cls(contract_hash=types.UInt160.from_string(value[2:]))
        if len(value) == 66:
            ecpoint = cryptography.ECPoint.deserialize_from_bytes(binascii.unhexlify(value))
            return cls(group=ecpoint)
        if value == '*':
            return cls()  # no args == wildcard
        raise ValueError(f"Invalid JSON - Cannot deduce permission type from: {value}")

    def to_array(self) -> bytes:
        """ Serialize the object """
        if self.is_hash:
            return self.contract_hash.to_array()  # type: ignore
        if self.is_group:
            return self.group.to_array()  # type: ignore
        # wildcard
        return b''



from .empty import EmptyPayload
from .inventory import InventoryPayload, InventoryType, IInventory
from .version import VersionPayload
from .address import NetworkAddress, AddrPayload, AddressState, DisconnectReason
from .ping import PingPayload
from .verification import Witness, WitnessScope, Signer, IVerifiable
from .transaction import Transaction, TransactionAttribute, TransactionAttributeType, HighPriorityAttribute
from .block import (Header,
                    Block,
                    MerkleBlockPayload,
                    HeadersPayload,
                    TrimmedBlock,
                    GetBlocksPayload,
                    GetBlockByIndexPayload)
from .filter import FilterAddPayload, FilterLoadPayload
from .oracle import OracleReponseCode, OracleResponse
from .extensible import ExtensiblePayload

__all__ = ['EmptyPayload', 'InventoryPayload', 'InventoryType', 'VersionPayload', 'NetworkAddress',
           'AddrPayload', 'PingPayload', 'Witness', 'WitnessScope', 'Header', 'Block', 'MerkleBlockPayload',
           'HeadersPayload', 'Transaction', 'TransactionAttribute',
           'TransactionAttributeType', 'Signer', 'GetBlocksPayload', 'GetBlockByIndexPayload', 'FilterAddPayload',
           'FilterLoadPayload', 'TrimmedBlock', 'IVerifiable', 'OracleReponseCode', 'OracleReponseCode',
           'ExtensiblePayload', 'HighPriorityAttribute']



from __future__ import annotations
import hashlib
from typing import List
from neo3 import storage, settings
from neo3.core import types, serialization, Size as s, utils
from neo3.network import payloads, message
from neo3.network.payloads import InventoryType


class ExtensiblePayload(payloads.IInventory):
    def __init__(self,
                 category: str,
                 valid_block_start: int,
                 valid_block_end: int,
                 sender: types.UInt160,
                 data: bytes,
                 witness: payloads.Witness):
        #: An identifier to which category the data belongs
        self.category = category
        #: Starting height in which the payload is valid
        self.valid_block_start = valid_block_start
        #: Last height height in which the payload is valid
        self.valid_block_end = valid_block_end
        #: The hash of the account who has send the payload to the network
        self.sender = sender
        #: Arbitrary data as required by the payload category
        self.data = data
        #: The witness of the payload
        self.witness = witness

    def __len__(self):
        return (utils.get_var_size(self.category)
                + s.uint32
                + s.uint32
                + s.uint160
                + utils.get_var_size(self.data)
                + 1
                + len(self.witness))

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_uint8(1)
        writer.write_serializable(self.witness)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned part of the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_string(self.category)
        writer.write_uint32(self.valid_block_start)
        writer.write_uint32(self.valid_block_end)
        writer.write_serializable(self.sender)
        writer.write_var_bytes(self.data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the check byte is not 1.
        """
        self.deserialize_unsigned(reader)
        if reader.read_uint8() != 1:
            raise ValueError("Deserialization error - check byte incorrect")
        self.witness = reader.read_serializable(payloads.Witness)

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the unsigned data part of the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the valid_block_start exceeds the valid_block_end field.
        """
        self.category = reader.read_var_string(32)
        self.valid_block_start = reader.read_uint32()
        self.valid_block_end = reader.read_uint32()
        if self.valid_block_start >= self.valid_block_end:
            raise ValueError("Deserialization error - valid_block_start is bigger than valid_block_end")
        self.sender = reader.read_serializable(types.UInt160)
        self.data = reader.read_var_bytes(message.Message.PAYLOAD_MAX_SIZE)

    def hash(self) -> types.UInt256:
        """
        Get a unique identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(data_to_hash).digest()
            return types.UInt256(data=data)

    @property
    def inventory_type(self) -> InventoryType:
        """
        Inventory type identifier.
        """
        return InventoryType.EXTENSIBLE

    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        return [self.sender]

    @classmethod
    def _serializable_init(cls):
        return cls('', 0, 0, types.UInt160.zero(), b'', payloads.Witness(b'', b''))



import asynctest
import socket
import logging
import asyncio
import binascii
from neo3.network import node, message, payloads, capabilities, ipfilter, protocol, encode_base62

from neo3 import settings, network_logger
from neo3.core import types
from unittest import mock
from tests import helpers as test_helpers


class NeoNodeSocketMock(asynctest.SocketMock):
    def __init__(self, loop, hostaddr: str, port: int):
        super(NeoNodeSocketMock, self).__init__()
        self.type = socket.SOCK_STREAM
        self.recv_buffer = bytearray()
        self.loop = loop
        self.hostaddr = hostaddr
        self.port = port
        self.recv_data = self._recv_data()
        caps = [capabilities.FullNodeCapability(0),
                capabilities.ServerCapability(n_type=capabilities.NodeCapabilityType.TCPSERVER, port=10333)]
        self.m_send_version = message.Message(msg_type=message.MessageType.VERSION,
                                              payload=payloads.VersionPayload(nonce=123,
                                                                              user_agent="NEO3-MOCK-CLIENT",
                                                                              capabilities=caps))
        self.m_verack = message.Message(msg_type=message.MessageType.VERACK)

    def _recv_data(self):
        yield self.m_send_version.to_array()
        yield self.m_verack.to_array()
        raise BlockingIOError
        # while True:
        #     yield message.Message(msg_type=message.MessageType.PING,
        #                           payload=payloads.PingPayload(height=0)).to_array()

    def recv(self, max_bytes):
        if not self.recv_buffer:
            try:
                self.recv_buffer.extend(next(self.recv_data))
                asynctest.set_read_ready(self, self.loop)
            except StopIteration:
                # nothing left
                pass

        data = self.recv_buffer[:max_bytes]
        self.recv_buffer = self.recv_buffer[max_bytes:]

        if self.recv_buffer:
            # Some more data to read
            asynctest.set_read_ready(self, self.loop)

        return data

    def send(self, data):
        asynctest.set_read_ready(self, self.loop)
        return len(data)

    def getpeername(self):
        return (self.hostaddr, self.port)


class NeoNodeTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        network_logger = logging.getLogger('neo3.network')
        network_logger.setLevel(logging.DEBUG)
        stdio_handler = logging.StreamHandler()
        network_logger.addHandler(stdio_handler)

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def setUp(self) -> None:
        settings.reset_settings_to_default()
        node.NeoNode._reset_for_test()
        ipfilter.ipfilter.reset()

    async def test_connect_to(self):
        with self.assertRaises(ValueError) as context:
            await node.NeoNode.connect_to()
        self.assertIn("host and port was not specified and no sock specified", str(context.exception))

        with self.assertRaises(ValueError) as context:
            await node.NeoNode.connect_to(host='123', socket=object())
        self.assertIn("host/port and socket can not be specified at the same time", str(context.exception))

    async def test_connect_to_with_socket(self):
        settings.network.magic = 769

        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)

        with self.assertLogs(network_logger, 'DEBUG') as log_context:
           n, _ = await node.NeoNode.connect_to(socket=socket_mock)
        self.assertIn("Trying to connect to socket", log_context.output[0])
        self.assertIn("Connected to NEO3-MOCK-CLIENT @ 127.0.0.1:1111", log_context.output[2])
        self.assertIsInstance(n, node.NeoNode)
        await n.disconnect(payloads.DisconnectReason.SHUTTING_DOWN)

    async def test_connect_to_with_host_ip(self):
        # just for basic coverage, we use the above socket mock for more indepth testing
        host = '127.0.0.1'
        port = 1111

        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with asynctest.patch.object(self.loop, 'create_connection'):
                _, _ = await node.NeoNode.connect_to(host, port)
        self.assertIn(f"Trying to connect to: {host}:{port}", log_context.output[0])

    async def test_connect_to_exceptions(self):
        loop = asynctest.MagicMock()

        loop.create_connection.side_effect = asyncio.TimeoutError
        node_instance, failure = await node.NeoNode.connect_to(socket=object(), loop=loop)
        self.assertIsNone(node_instance)
        self.assertEqual("Timed out", failure[1])

        loop.create_connection.side_effect = OSError("unreachable")
        node_instance, failure = await node.NeoNode.connect_to(socket=object(), loop=loop)
        self.assertIsNone(node_instance)
        self.assertEqual("Failed to connect for reason unreachable", failure[1])

        loop.create_connection.side_effect = asyncio.CancelledError
        node_instance, failure = await node.NeoNode.connect_to(socket=object(), loop=loop)
        self.assertIsNone(node_instance)
        self.assertEqual("Cancelled", failure[1])

    def test_helpers(self):
        addr1 = payloads.NetworkAddress(address='127.0.0.1:1111')
        addr1.set_state_dead()
        addr2 = payloads.NetworkAddress(address='127.0.0.1:2222')
        node.NeoNode.addresses = [addr1, addr2]
        result = node.NeoNode.get_address_new()
        self.assertEqual(addr2, result)

        n = node.NeoNode(object())
        addr = n._find_address_by_host_port('127.0.0.1:1111')
        self.assertEqual(addr1, addr)
        addr = n._find_address_by_host_port('127.0.0.1:3333')
        self.assertEqual(None, addr)

    async def test_handshake_first_message_not_VERSION(self):
        settings.network.magic = 769
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)

        def _recv_data2(self):
            yield self.m_verack.to_array()

        socket_mock.recv_data = _recv_data2(socket_mock)
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            await node.NeoNode.connect_to(socket=socket_mock)
        self.assertIn("Disconnect called with reason=HANDSHAKE_VERSION_ERROR", log_context.output[1])

    async def test_handshake_version_validation_failed(self):
        # mock the version validation call result. The version_validation function is tested down below
        settings.network.magic = 769
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with asynctest.patch('neo3.network.node.NeoNode._validate_version', return_value=False):
                await node.NeoNode.connect_to(socket=socket_mock)

        self.assertIn("Disconnect called with reason=HANDSHAKE_VERSION_ERROR", log_context.output[1])

    async def test_handshake_second_message_not_VERACK(self):
        settings.network.magic = 769
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)

        def _recv_data2(self):
            yield self.m_send_version.to_array()
            yield self.m_send_version.to_array()

        socket_mock.recv_data = _recv_data2(socket_mock)

        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            await node.NeoNode.connect_to(socket=socket_mock)
        self.assertIn("Disconnect called with reason=HANDSHAKE_VERACK_ERROR", log_context.output[2])

    def _new_version(self):
        self.start_height = 888
        caps = [capabilities.FullNodeCapability(start_height=self.start_height),
                capabilities.ServerCapability(capabilities.NodeCapabilityType.TCPSERVER, 10333)]
        return payloads.VersionPayload(nonce=123, user_agent="NEO3-MOCK-CLIENT", capabilities=caps)

    def test_version_validation_client_is_self(self):
        n = node.NeoNode(object())

        # test for client is self
        version = self._new_version()
        version.nonce = n.nodeid
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            result = n._validate_version(version)
        self.assertFalse(result)
        self.assertIn("Client is self", log_context.output[0])

    def test_version_validation_wrong_network(self):
        n = node.NeoNode(object())

        # test for wrong network
        settings.network.magic = 769
        version = self._new_version()
        version.magic = 111
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            result = n._validate_version(version)
        self.assertFalse(result)
        self.assertIn("Wrong network id", log_context.output[0])

    def test_version_validation_should_updating_address_to_connected_state(self):
        n = node.NeoNode(object())

        # test updating address state to CONNECTED
        # this is relevant for addresses that have been added through the nodemanager based on the seedlist
        # set the addresses ourselves to mimick the nodemanager startup behaviour
        node.NeoNode.addresses = [n.address]
        version = self._new_version()

        result = n._validate_version(version)
        # validate our address is now set to CONNECTED
        self.assertTrue(result)
        self.assertTrue(n.address.is_state_connected)
        self.assertEqual(self.start_height, n.best_height)

    def test_version_validation_should_add_new_address(self):
        # when using the `connect_to` method or when a server is hosted accepting incoming clients
        # we should add the address to our know addresses list
        n = node.NeoNode(object())
        # normally this is updated by `connection_made()`, since we skip that in this test we set it manually
        n.address.address = '127.0.0.1:1111'
        self.assertEqual(0, len(node.NeoNode.addresses))

        version = self._new_version()
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            n._validate_version(version)
        self.assertEqual(1, len(node.NeoNode.addresses))
        self.assertIn("Adding address from outside 127.0.0.1:1111", log_context.output[0])

    def test_version_validation_fail_if_no_full_node_capabilities(self):
        n = node.NeoNode(object())

        version = self._new_version()
        # remove the full node capability
        version.capabilities.pop(0)
        result = n._validate_version(version)
        self.assertFalse(result)

    async def test_connect_blocked_by_ipfilter(self):
        settings.network.magic = 769
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)
        ipfilter.ipfilter.blacklist_add('127.0.0.1')

        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            await node.NeoNode.connect_to(socket=socket_mock)
        self.assertIn("Blocked by ipfilter: 127.0.0.1:1111", log_context.output[1])

    async def test_req_addr_list(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        await n.request_address_list()

        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETADDR, m.type)
        self.assertIsInstance(m.payload, payloads.EmptyPayload)

    async def test_send_addr_list(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        n.addresses = [payloads.NetworkAddress(address='127.0.0.1:1111')]
        await n.send_address_list(n.addresses)

        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.ADDR, m.type)
        self.assertIsInstance(m.payload, payloads.AddrPayload)
        self.assertEqual(n.addresses, m.payload.addresses)

    async def test_req_headers(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        index_start = 0
        count = 10

        await n.request_headers(index_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETHEADERS, m.type)
        self.assertIsInstance(m.payload, payloads.GetBlockByIndexPayload)
        self.assertEqual(index_start, m.payload.index_start)
        self.assertEqual(count, m.payload.count)

    async def test_send_headers(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        headers = 2001 * [test_helpers.SerializableObject()]

        await n.send_headers(headers)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.HEADERS, m.type)
        self.assertIsInstance(m.payload, payloads.HeadersPayload)
        # max sure it clips the size to max 2K headers
        self.assertEqual(2000, len(m.payload.headers))

    async def test_req_blocks(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        hash_start = types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9")
        count = 10

        await n.request_blocks(hash_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETBLOCKS, m.type)
        self.assertIsInstance(m.payload, payloads.GetBlocksPayload)
        self.assertEqual(hash_start, m.payload.hash_start)
        self.assertEqual(count, m.payload.count)

    async def test_req_block_data(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        index_start = 1
        count = 2

        await n.request_block_data(index_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETBLOCKBYINDEX, m.type)
        self.assertIsInstance(m.payload, payloads.GetBlockByIndexPayload)
        self.assertEqual(index_start, m.payload.index_start)
        self.assertEqual(count, m.payload.count)

    async def test_request_data(self):
        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()
        hash1 = types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9")
        hash2 = types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aaaaa")
        hashes = [hash1, hash2]

        await n.request_data(payloads.InventoryType.BLOCK, hashes)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETDATA, m.type)
        self.assertIsInstance(m.payload, payloads.InventoryPayload)
        self.assertEqual(hashes, m.payload.hashes)

    async def test_send_inventory_and_relay(self):
        # test 2 in 1
        # taken from the Transaction testcase in `test_payloads.py`
        raw_tx = binascii.unhexlify(b'007B000000C8010000000000001503000000000000010000000154A64CAC1B1073E662933EF3E30B007CD98D67D7000002010201000155')
        tx = payloads.Transaction.deserialize_from_bytes(raw_tx)

        n = node.NeoNode(object())
        n.send_message = asynctest.CoroutineMock()

        await n.relay(tx)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.INV, m.type)
        self.assertIsInstance(m.payload, payloads.InventoryPayload)
        self.assertEqual(tx.hash(), m.payload.hashes[0])

    async def test_processing_messages(self):
        settings.network.magic = 769
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)

        m_addr = message.Message(msg_type=message.MessageType.ADDR, payload=payloads.AddrPayload([]))
        m_block = message.Message(msg_type=message.MessageType.BLOCK, payload=payloads.EmptyPayload())
        m_inv1 = message.Message(msg_type=message.MessageType.INV, payload=payloads.InventoryPayload(
            payloads.InventoryType.BLOCK,
            hashes=[types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9")]
        ))
        m_inv2 = message.Message(msg_type=message.MessageType.INV,
                                 payload=payloads.InventoryPayload(payloads.InventoryType.TX, []))
        m_getaddr = message.Message(msg_type=message.MessageType.GETADDR, payload=payloads.EmptyPayload())
        m_mempool = message.Message(msg_type=message.MessageType.MEMPOOL, payload=payloads.EmptyPayload())

        # taken from the Headers testcase in `test_payloads`
        raw_headers_payload = binascii.unhexlify(b'0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A400000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E201020102020304')
        m_headers = message.Message(msg_type=message.MessageType.HEADERS,
                                    payload=payloads.HeadersPayload.deserialize_from_bytes(raw_headers_payload))
        m_ping = message.Message(msg_type=message.MessageType.PING, payload=payloads.PingPayload(0))
        m_pong = message.Message(msg_type=message.MessageType.PONG, payload=payloads.PingPayload(0))
        m_reject = message.Message(msg_type=message.MessageType.REJECT, payload=payloads.EmptyPayload())

        def _recv_data2(self):
            # first do handshake
            yield self.m_send_version.to_array()
            yield self.m_verack.to_array()
            # next send all types of messages we handle
            yield m_addr.to_array()
            yield m_block.to_array()
            yield m_inv1.to_array()
            yield m_inv2.to_array()
            yield m_getaddr.to_array()
            yield m_mempool.to_array()
            yield m_headers.to_array()
            yield m_ping.to_array()
            yield m_pong.to_array()
            yield m_reject.to_array()

        socket_mock.recv_data = _recv_data2(socket_mock)
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            try:
                n, _ = await node.NeoNode.connect_to(socket=socket_mock)
            except Exception as e:
                print(f"GVD {e}")

            await asyncio.sleep(0.5)
            await n.disconnect(payloads.DisconnectReason.SHUTTING_DOWN)

    @asynctest.SkipTest
    async def test_processing_messages3(self):
        # we got 2 cases for which we need to test without using a backend
        settings.network.magic = 769
        settings.storage.use_default = False
        socket_mock = NeoNodeSocketMock(self.loop, '127.0.0.1', 1111)

        m_inv1 = message.Message(msg_type=message.MessageType.INV, payload=payloads.InventoryPayload(
            payloads.InventoryType.BLOCK,
            hashes=[types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9")]
        ))
        m_ping = message.Message(msg_type=message.MessageType.PING, payload=payloads.PingPayload(0))

        def _recv_data2(self):
            print("my recv data 2 called ")
            # first do handshake
            yield self.m_send_version.to_array()
            yield self.m_verack.to_array()
            yield m_inv1.to_array()
            yield m_ping.to_array()

        socket_mock.recv_data = _recv_data2(socket_mock)
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with mock.patch('neo3.network.node.NeoNode.send_message', new_callable=asynctest.CoroutineMock):
            # with asynctest.patch('neo3.network.node.NeoNode.send_message', return_value=asynctest.CoroutineMock()):
                n, _ = await node.NeoNode.connect_to(socket=socket_mock)
                await asyncio.sleep(0.1)
                await n.disconnect(payloads.DisconnectReason.SHUTTING_DOWN)
        # print(n.send_message.call_args)

    # async def test_wtf(self):
    #     protocol = asynctest.MagicMock()
    #     protocol.send_message = asynctest.CoroutineMock()
    #     nn = node.NeoNode(protocol)
    #     mock_read_message = asynctest.CoroutineMock()
    #     nn.read_message = mock_read_message
    #
    #     m_inv1 = message.Message(msg_type=message.MessageType.INV, payload=payloads.InventoryPayload(
    #         payloads.InventoryType.BLOCK,
    #         hashes=[types.UInt256.from_string("65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9")]
    #     ))
    #     m_ping = message.Message(msg_type=message.MessageType.PING, payload=payloads.PingPayload(0))
    #
    #     mock_read_message.side_effect = [m_inv1, m_ping]
    #     await nn._process_incoming_data()

    def test_utility_function(self):
        with self.assertRaises(ValueError) as context:
            encode_base62(-100)
        self.assertIn("cannot encode negative numbers", str(context.exception))

        self.assertEqual('0', encode_base62(0))

        self.assertEqual('w1R', encode_base62(123123))




import unittest
from neo3 import vm, contracts, storage, settings
from neo3.core import to_script_hash
from neo3.contracts import interop
from tests.contracts.interop.utils import syscall_name_to_int, test_engine

"""
In the early days, when all these tests were written, contract hashes were determined by hashing the contract script.
This means that ApplicationEngine.calling_script_hash equaled a hash of the script we loaded via load_script(). 
That knowledge was used initially for creating all the tests below. 
 
The contract hash computation changed with:
- https://github.com/neo-project/neo/pull/2244
- https://github.com/neo-project/neo/pull/2240

Because of this change some tricks are used internally by NEO for updating the `calling_script_hash` to the correct contract hash. 
These tricks only happen when calling contracts via `System.Contract.Call` or via the OpCode.CALLT. 
Using these make the tests far more complex, so instead we keep the old logic. \

In summary this means; 
ContractState objects are created where the ContractHash is still based on the old method of hashing
the contract script. Technically this is incorrect, but makes testing much easier.
"""


class StorageInteropTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    @classmethod
    def setUpClass(cls) -> None:
        # blockchain.Blockchain()
        pass

    def setUp(self) -> None:
        self.RET = b'\x40'
        self.manifest = contracts.ContractManifest("contract_name")
        self.nef = contracts.NEF(script=self.RET)
        self.contract_hash = to_script_hash(self.nef.script)
        self.contract = contracts.ContractState(1, self.nef, self.manifest, 0, self.contract_hash)
        self.contract.manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]

    def test_get_context(self):
        engine = test_engine(has_snapshot=True)
        engine.snapshot.contracts.put(self.contract)
        ctx = engine.invoke_syscall_by_name("System.Storage.GetContext")
        self.assertIsInstance(ctx, storage.StorageContext)
        self.assertEqual(self.contract.id, ctx.id)
        self.assertFalse(ctx.is_read_only)

        # repeat for read only
        ctx = engine.invoke_syscall_by_name("System.Storage.GetReadOnlyContext")
        self.assertIsInstance(ctx, storage.StorageContext)
        self.assertEqual(self.contract.id, ctx.id)
        self.assertTrue(ctx.is_read_only)

    def test_as_readonly(self):
        engine = test_engine(has_snapshot=True)
        engine.snapshot.contracts.put(self.contract)
        ctx = engine.invoke_syscall_by_name("System.Storage.GetContext")
        self.assertFalse(ctx.is_read_only)
        engine.push(vm.StackItem.from_interface(ctx))
        ctx = engine.invoke_syscall_by_name("System.Storage.AsReadOnly")
        self.assertTrue(ctx.is_read_only)

    def test_storage_get_key_not_found(self):
        engine = test_engine(has_snapshot=True, has_container=True)
        script = vm.ScriptBuilder()
        # key parameter for the `Storage.Get` syscall
        script.emit(vm.OpCode.PUSH2)
        script.emit_syscall(syscall_name_to_int("System.Storage.GetContext"))
        # at this point our stack looks like follows
        # * storage context
        # * key
        script.emit_syscall(syscall_name_to_int("System.Storage.Get"))
        engine.load_script(vm.Script(script.to_array()))

        # we have to store our contract or some sanity checks will fail (like getting a StorageContext
        nef = contracts.NEF(script=script.to_array())
        contract = contracts.ContractState(1, nef, self.manifest, 0, to_script_hash(nef.script))
        engine.snapshot.contracts.put(contract)

        storage_key = storage.StorageKey(contract.id, b'\x01')
        storage_item = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key, storage_item)

        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertIsInstance(item, vm.NullStackItem)

    def test_storage_get_ok(self):
        engine = test_engine(has_snapshot=True)
        engine.snapshot.contracts.put(self.contract)

        storage_key = storage.StorageKey(self.contract.id, b'\x01')
        storage_item = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key, storage_item)

        ctx = engine.invoke_syscall_by_name("System.Storage.GetContext")
        engine.push(vm.ByteStringStackItem(storage_key.key))
        engine.push(vm.StackItem.from_interface(ctx))
        returned_value = engine.invoke_syscall_by_name("System.Storage.Get")

        self.assertEqual(storage_item.value, returned_value)

    def test_storage_get_ok2(self):
        # this is basically the same as `test_storage_get_ok`
        # but performed by executing a script
        # it exists to validate that the `Optional[bytes]` return value is converted properly
        engine = test_engine(has_snapshot=True)
        script = vm.ScriptBuilder()
        script.emit(vm.OpCode.PUSH1)
        script.emit_syscall(syscall_name_to_int("System.Storage.GetContext"))
        script.emit_syscall(syscall_name_to_int("System.Storage.Get"))
        engine.load_script(vm.Script(script.to_array()))

        nef = contracts.NEF(script=script.to_array())
        contract_hash = to_script_hash(nef.script)
        contract = contracts.ContractState(1, nef, self.manifest, 0, contract_hash)
        engine.snapshot.contracts.put(contract)

        storage_key = storage.StorageKey(contract.id, b'\x01')
        storage_item = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key, storage_item)

        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertEqual(storage_item.value, item.to_array())

    def test_storage_find(self):
        # settings.storage.default_provider = 'leveldb'
        engine = test_engine(has_snapshot=True)
        engine.snapshot.contracts.put(self.contract)

        storage_key1 = storage.StorageKey(self.contract.id, b'\x01')
        storage_item1 = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key1, storage_item1)
        storage_key2 = storage.StorageKey(self.contract.id, b'\x02')
        storage_item2 = storage.StorageItem(b'\x22')
        engine.snapshot.storages.put(storage_key2, storage_item2)

        ctx = engine.invoke_syscall_by_name("System.Storage.GetContext")
        engine.push(vm.IntegerStackItem(contracts.FindOptions.NONE))
        engine.push(vm.ByteStringStackItem(storage_key1.key))
        engine.push(vm.StackItem.from_interface(ctx))

        it = engine.invoke_syscall_by_name("System.Storage.Find")
        self.assertIsInstance(it, interop.StorageIterator)

        with self.assertRaises(ValueError) as context:
            it.value()
        self.assertEqual("Cannot call 'value' without having advanced the iterator at least once", str(context.exception))

        self.assertTrue(it.next())

        struct = it.value()  # 0 key, 1 value
        self.assertEqual(storage_item1.value, struct[1].to_array())

    def test_storage_put_helper_parameter_validation(self):
        with self.assertRaises(ValueError) as context:
            key = (b'\x01' * contracts.interop.MAX_STORAGE_KEY_SIZE) + b'\x01'
            contracts.interop.storage_put(None, None, key, b'')
        self.assertEqual(f"Storage key length exceeds maximum of {contracts.interop.MAX_STORAGE_KEY_SIZE}", str(context.exception))

        with self.assertRaises(ValueError) as context:
            value = (b'\x01' * contracts.interop.MAX_STORAGE_VALUE_SIZE) + b'\x01'
            contracts.interop.storage_put(None, None, b'', value)
        self.assertEqual(f"Storage value length exceeds maximum of {contracts.interop.MAX_STORAGE_VALUE_SIZE}", str(context.exception))

        with self.assertRaises(ValueError) as context:
            ctx = storage.StorageContext(None, is_read_only=True)
            contracts.interop.storage_put(None, ctx, b'', b'')
        self.assertEqual("Cannot persist to read-only storage context", str(context.exception))

    def test_storage_put_new(self):
        # see `test_storage_get_key_not_found()` for a description on why the storage is setup with a script as is

        # setup
        engine = test_engine(has_snapshot=True)
        script = vm.ScriptBuilder()
        script.emit(vm.OpCode.PUSH2)  # storage put value
        script.emit(vm.OpCode.PUSH1)  # storage put key
        script.emit_syscall(syscall_name_to_int("System.Storage.GetContext"))
        script.emit_syscall(syscall_name_to_int("System.Storage.Put"))
        engine.load_script(vm.Script(script.to_array()))

        nef = contracts.NEF(script=script.to_array())
        manifest = contracts.ContractManifest(f"contractname1")
        manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]
        hash_ = to_script_hash(nef.script)

        contract = contracts.ContractState(1, nef, manifest, 0, hash_)
        engine.snapshot.contracts.put(contract)

        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        storage_key = storage.StorageKey(1, b'\x01')
        item = engine.snapshot.storages.try_get(storage_key)
        self.assertIsNotNone(item)
        self.assertEqual(b'\x02', item.value)

    def test_storage_put_overwrite(self):
        # test with new data being shorter than the old data
        engine = test_engine(has_snapshot=True)
        key = b'\x01'
        storage_key = storage.StorageKey(1, key)
        storage_item = storage.StorageItem(b'\x11\x22\x33')
        engine.snapshot.storages.put(storage_key, storage_item)

        ctx = storage.StorageContext(1, is_read_only=False)
        new_item_value = b'\x11\x22'
        contracts.interop.storage_put(engine, ctx, key, new_item_value)

        item = engine.snapshot.storages.get(storage_key)
        self.assertIsNotNone(item)
        self.assertEqual(new_item_value, item.value)

        # now test with data being longer than before
        longer_item_value = b'\x11\x22\x33\x44'
        contracts.interop.storage_put(engine, ctx, key, longer_item_value)

        item = engine.snapshot.storages.get(storage_key)
        self.assertIsNotNone(item)
        self.assertEqual(longer_item_value, item.value)

    def test_storage_delete_readonly_context(self):
        engine = test_engine(has_snapshot=True)

        engine.snapshot.contracts.put(self.contract)

        storage_key = storage.StorageKey(self.contract.id, b'\x01')
        storage_item = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key, storage_item)

        ctx = engine.invoke_syscall_by_name("System.Storage.GetReadOnlyContext")
        engine.push(vm.ByteStringStackItem(storage_key.key))
        engine.push(vm.StackItem.from_interface(ctx))

        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Storage.Delete")
        self.assertEqual("Cannot delete from read-only storage context", str(context.exception))

    def test_delete_ok(self):
        engine = test_engine(has_snapshot=True)
        engine.snapshot.contracts.put(self.contract)

        storage_key = storage.StorageKey(self.contract.id, b'\x01')
        storage_item = storage.StorageItem(b'\x11')
        engine.snapshot.storages.put(storage_key, storage_item)

        ctx = engine.invoke_syscall_by_name("System.Storage.GetContext")
        engine.push(vm.ByteStringStackItem(storage_key.key))
        engine.push(vm.StackItem.from_interface(ctx))

        engine.invoke_syscall_by_name("System.Storage.Delete")

        self.assertIsNone(engine.snapshot.storages.try_get(storage_key))



from __future__ import annotations
from neo3.core import serialization
from neo3crypto import (ECPoint as _ECPointCpp,  # type: ignore
                        ECCCurve,
                        ECCException,
                        sign as ecdsa_sign,
                        verify as ecdsa_verify)
from typing import Type, Union, Any
import os
import binascii


# mypy workaround
type_ECPoint = type(_ECPointCpp)  # type: Any
type_Serializable = type(serialization.ISerializable)  # type: Any


class SerializableECPointMeta(type_ECPoint, type_Serializable):
    pass


class ECPoint(_ECPointCpp, serialization.ISerializable, metaclass=SerializableECPointMeta):
    def __init__(self, *args, **kwargs):
        super(ECPoint, self).__init__(*args, **kwargs)

    def __str__(self):
        return binascii.hexlify(self.encode_point(compressed=True)).decode('utf8')

    def __bool__(self):
        return True

    def __hash__(self):
        return hash(self.x + self.y)

    def __deepcopy__(self, memodict={}):
        return ECPoint.deserialize_from_bytes(self.to_array(), self.curve, False)

    def is_zero(self):
        return self.x == 0 and self.y == 0

    def serialize(self, writer: serialization.BinaryWriter, compress=True) -> None:
        if self.is_infinity:
            writer.write_bytes(b'\x00')
        else:
            writer.write_bytes(self.encode_point(compress))

    def deserialize(self, reader: serialization.BinaryReader, curve=ECCCurve.SECP256R1) -> None:
        try:
            f0 = reader.read_byte()
        except ValueError:
            # infinity
            self.from_bytes(b'\x00', curve, True)
            return

        f1 = int.from_bytes(f0, 'little')
        if f1 == 0:
            # infinity
            self.from_bytes(b'\x00', curve, True)
            return
        elif f1 == 2 or f1 == 3:
            data = reader.read_bytes(32)
            self.from_bytes(f0 + data, curve, True)
            return
        else:
            raise ValueError(f"Unsupported point encoding: {str(f0)}")

    @classmethod
    def deserialize_from_bytes(cls: Type[serialization.ISerializable_T],
                               data: Union[bytes, bytearray],
                               curve: ECCCurve = ECCCurve.SECP256R1,
                               validate: bool = True) -> serialization.ISerializable_T:
        """
        Parse data into an object instance.

        Args:
            data: ECPoint in hex escaped bytes format.
            curve: the curve type to decompress
            validate: validate if the point valid point on the specified curve

        Returns:
            a deserialized instance of the class.
        """
        return cls(data, curve, validate)  # type: ignore

    @classmethod
    def _serializable_init(cls):
        return cls(b'\x00', ECCCurve.SECP256R1, False)


class KeyPair:
    def __init__(self, private_key: bytes, curve: ECCCurve = ECCCurve.SECP256R1):
        self.private_key = private_key
        self.public_key: ECPoint = ECPoint(private_key, curve)

    @classmethod
    def generate(cls):
        return cls(os.urandom(32))



from __future__ import annotations
from . import NativeContract, register
from neo3.core import types
from neo3 import storage, contracts, vm


class PolicyContract(NativeContract):
    _id: int = -7

    DEFAULT_EXEC_FEE_FACTOR = 30
    MAX_EXEC_FEE_FACTOR = 100
    DEFAULT_FEE_PER_BYTE = 1000
    DEFAULT_STORAGE_PRICE = 100000
    MAX_STORAGE_PRICE = 10000000

    key_fee_per_byte = storage.StorageKey(_id, b'\x0A')
    key_blocked_account = storage.StorageKey(_id, b'\x0F')
    key_exec_fee_factor = storage.StorageKey(_id, b'\x12')
    key_storage_price = storage.StorageKey(_id, b'\x13')

    _storage_price = 0

    def init(self):
        super(PolicyContract, self).init()

    @register("getFeePerByte", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_fee_per_byte(self, snapshot: storage.Snapshot) -> int:
        """
        Retrieve the configured maximum fee per byte of storage.

        Returns:
            int: maximum fee.
        """
        data = snapshot.storages.get(self.key_fee_per_byte, read_only=True)
        return int.from_bytes(data.value, 'little', signed=True)

    @register("isBlocked", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def is_blocked(self, snapshot: storage.Snapshot, account: types.UInt160) -> bool:
        """
        Check if the account is blocked

        Transaction from blocked accounts will be rejected by the consensus nodes.
        """

        si = snapshot.storages.try_get(self.key_blocked_account + account.to_array(), read_only=True)
        if si is None:
            return False
        else:
            return True

    @register("setFeePerByte", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _set_fee_per_byte(self, engine: contracts.ApplicationEngine, value: int) -> None:
        """
        Should only be called through syscalls
        """
        if value < 0 or value > 100000000:
            raise ValueError("New value exceeds FEE_PER_BYTE limits")

        if not self._check_committee(engine):
            raise ValueError("Check committee failed")

        storage_item = engine.snapshot.storages.get(self.key_fee_per_byte, read_only=False)
        storage_item.value = self._int_to_bytes(value)

    @register("blockAccount", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _block_account(self, engine: contracts.ApplicationEngine, account: types.UInt160) -> bool:
        """
        Should only be called through syscalls
        """
        if not self._check_committee(engine):
            return False
        if self.is_native(account):
            raise ValueError("Cannot block native contracts")
        storage_key = self.key_blocked_account + account.to_array()
        storage_item = engine.snapshot.storages.try_get(storage_key, read_only=False)
        if storage_item is None:
            storage_item = storage.StorageItem(b'\x00')
            engine.snapshot.storages.update(storage_key, storage_item)
        else:
            return False

        return True

    @register("unblockAccount", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _unblock_account(self, engine: contracts.ApplicationEngine, account: types.UInt160) -> bool:
        """
        Should only be called through syscalls
        """
        if not self._check_committee(engine):
            return False
        storage_key = self.key_blocked_account + account.to_array()
        storage_item = engine.snapshot.storages.try_get(storage_key, read_only=False)
        if storage_item is None:
            return False
        else:
            engine.snapshot.storages.delete(storage_key)
        return True

    @register("getExecFeeFactor", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_exec_fee_factor(self, snapshot: storage.Snapshot) -> int:
        storage_item = snapshot.storages.get(self.key_exec_fee_factor, read_only=True)
        return int(vm.BigInteger(storage_item.value))

    @register("getStoragePrice", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_storage_price(self, snapshot: storage.Snapshot) -> int:
        if self._storage_price:
            return self._storage_price

        storage_item = snapshot.storages.get(self.key_storage_price, read_only=True)
        return int(vm.BigInteger(storage_item.value))

    @register("setExecFeeFactor", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _set_exec_fee_factor(self, engine: contracts.ApplicationEngine, value: int) -> None:
        if value == 0 or value > self.MAX_EXEC_FEE_FACTOR:
            raise ValueError("New exec fee value out of range")
        if not self._check_committee(engine):
            raise ValueError("Check committee failed")
        storage_item = engine.snapshot.storages.get(self.key_exec_fee_factor, read_only=False)
        storage_item.value = vm.BigInteger(value).to_array()

    @register("setStoragePrice", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _set_storage_price(self, engine: contracts.ApplicationEngine, value: int) -> None:
        if value == 0 or value > self.MAX_STORAGE_PRICE:
            raise ValueError("New storage price value out of range")
        if not self._check_committee(engine):
            raise ValueError("Check committee failed")
        storage_item = engine.snapshot.storages.get(self.key_storage_price, read_only=False)
        storage_item.value = vm.BigInteger(value).to_array()

        self._storage_price = value

    def _int_to_bytes(self, value: int) -> bytes:
        return value.to_bytes((value.bit_length() + 7 + 1) // 8, 'little', signed=True)  # +1 for signed

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        def _to_si(value: int) -> storage.StorageItem:
            return storage.StorageItem(self._int_to_bytes(value))

        engine.snapshot.storages.put(self.key_fee_per_byte, _to_si(self.DEFAULT_FEE_PER_BYTE))
        engine.snapshot.storages.put(self.key_exec_fee_factor, _to_si(self.DEFAULT_EXEC_FEE_FACTOR))
        engine.snapshot.storages.put(self.key_storage_price, _to_si(self.DEFAULT_STORAGE_PRICE))



from enum import IntFlag


class FindOptions(IntFlag):
    NONE = 0
    KEYS_ONLY = 1 << 0
    REMOVE_PREFIX = 1 << 1
    VALUES_ONLY = 1 << 2
    DESERIALIZE_VALUES = 1 << 3
    PICK_FIELD0 = 1 << 4
    PICK_FIELD1 = 1 << 5
    ALL = KEYS_ONLY | REMOVE_PREFIX | VALUES_ONLY | DESERIALIZE_VALUES | PICK_FIELD0 | PICK_FIELD1



import json
import os.path
import unittest

from neo3 import wallet
from neo3.wallet import nep6
from neo3.wallet.wallet import Wallet


class WalletCreationTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_wallet_new_wallet(self):
        wallet_file_name = 'unittest-wallet'
        wallet_file_path = f"{wallet_file_name}.json"

        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_file_path):
            os.remove(wallet_file_path)

        test_wallet = nep6.NEP6DiskWallet(wallet_file_name)
        scrypt_parameters_default = wallet.ScryptParameters()

        self.assertEqual(wallet_file_name, test_wallet.name)
        self.assertEqual('1.0', test_wallet.version)
        self.assertEqual(scrypt_parameters_default.n, test_wallet.scrypt.n)
        self.assertEqual(scrypt_parameters_default.r, test_wallet.scrypt.r)
        self.assertEqual(scrypt_parameters_default.p, test_wallet.scrypt.p)
        self.assertEqual([], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)

    def test_wallet_default_value(self):
        test_wallet = nep6.NEP6DiskWallet.default()
        scrypt_parameters_default = wallet.ScryptParameters()

        self.assertEqual('wallet.json', test_wallet.name)
        self.assertEqual('1.0', test_wallet.version)
        self.assertEqual(scrypt_parameters_default.n, test_wallet.scrypt.n)
        self.assertEqual(scrypt_parameters_default.r, test_wallet.scrypt.r)
        self.assertEqual(scrypt_parameters_default.p, test_wallet.scrypt.p)
        self.assertEqual([], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)

    def test_wallet_save(self):
        wallet_path = 'unittest-wallet-save.json'
        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

        test_wallet = nep6.NEP6DiskWallet.default(wallet_path, 'NEP6 Wallet')
        test_wallet.save()
        self.assertTrue(os.path.isfile(wallet_path))

        with open(wallet_path) as json_file:
            data = json.load(json_file)

        self.assertEqual(data['name'], test_wallet.name)
        self.assertEqual(data['version'], test_wallet.version)
        self.assertEqual(data['scrypt']['n'], test_wallet.scrypt.n)
        self.assertEqual(data['scrypt']['r'], test_wallet.scrypt.r)
        self.assertEqual(data['scrypt']['p'], test_wallet.scrypt.p)
        self.assertEqual(data['accounts'], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)
        self.assertEqual(None, data['extra'])

        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

        # save using context manager
        with nep6.NEP6DiskWallet.default(wallet_path, 'NEP6 Wallet'):
            pass
        self.assertTrue(os.path.isfile(wallet_path))

        # clean up after test
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

    def test_wallet_from_json(self):
        password = '123'

        new_wallet = nep6.NEP6DiskWallet.default()
        test_account = wallet.Account.create_new(password)
        new_wallet.accounts.append(test_account)
        new_wallet._default_account = test_account

        json_wallet = new_wallet.to_json()

        test_wallet = Wallet.from_json(json_wallet, password='123')
        self.assertEqual(new_wallet.name, test_wallet.name)
        self.assertEqual('1.0', test_wallet.version)
        self.assertEqual(1, len(test_wallet.accounts))
        self.assertEqual(test_account, test_wallet.accounts[0])
        self.assertEqual(test_wallet._default_account, test_wallet.accounts[0])

    def test_wallet_account_new(self):
        password = 'abcabc'
        wallet = nep6.NEP6DiskWallet.default()
        self.assertEqual(0, len(wallet.accounts))

        # create account without label
        account = wallet.account_new(password)
        self.assertEqual(1, len(wallet.accounts))
        self.assertEqual(None, account.label)
        self.assertEqual(wallet._default_account, account)

        # create account with label
        label = 'New Account'
        account = wallet.account_new(password, label)
        self.assertEqual(2, len(wallet.accounts))
        self.assertEqual(label, account.label)
        self.assertNotEqual(wallet._default_account, account)

        # create account with duplicated label
        with self.assertRaises(ValueError):
            # label already used
            wallet.account_new(password, label)

        # create account and set as default
        label = 'Other Account'
        account = wallet.account_new(password, label, is_default=True)
        self.assertEqual(3, len(wallet.accounts))
        self.assertEqual(label, account.label)
        self.assertEqual(wallet._default_account, account)

    def test_wallet_account_add(self):
        password = 'abcabc'
        test_wallet = nep6.NEP6DiskWallet.default()
        self.assertEqual(0, len(test_wallet.accounts))

        label = 'New Account'
        account_1 = wallet.Account(password=password)
        account_2 = wallet.Account(password=password,
                                   label=label)
        account_3 = wallet.Account(password=password)
        account_4 = wallet.Account(password=password,
                                   label=label)

        # add account, first account is set as default
        success = test_wallet.account_add(account_1)
        self.assertTrue(success)
        self.assertEqual(1, len(test_wallet.accounts))
        self.assertEqual(test_wallet._default_account, account_1)

        # add account
        success = test_wallet.account_add(account_2)
        self.assertTrue(success)
        self.assertEqual(2, len(test_wallet.accounts))
        self.assertNotEqual(test_wallet._default_account, account_2)

        # add account already added
        success = test_wallet.account_add(account_2)
        self.assertFalse(success)

        # add account and set it as default
        success = test_wallet.account_add(account_3, is_default=True)
        self.assertTrue(success)
        self.assertEqual(3, len(test_wallet.accounts))
        self.assertEqual(test_wallet._default_account, account_3)

        # add account with duplicated label
        with self.assertRaises(ValueError):
            # label already used
            test_wallet.account_add(account_4)

    def test_wallet_account_delete(self):
        password = 'abcabc'
        account_1 = wallet.Account(password=password)
        account_2 = wallet.Account(password=password)
        account_3 = wallet.Account(password=password)

        test_wallet = nep6.NEP6DiskWallet.default()
        test_wallet.account_add(account_1)
        test_wallet.account_add(account_2)
        test_wallet.account_add(account_3)

        self.assertEqual(account_1, test_wallet._default_account)

        # delete account that is not default
        success = test_wallet.account_delete(account_2)
        self.assertTrue(success)
        self.assertEqual(account_1, test_wallet._default_account)

        # delete account not included
        success = test_wallet.account_delete(account_2)
        self.assertFalse(success)

        # delete account default, with other existing
        success = test_wallet.account_delete(account_1)
        self.assertTrue(success)
        self.assertNotEqual(account_1, test_wallet._default_account)
        self.assertEqual(account_3, test_wallet._default_account)

        # delete account default and it's the only existing account
        success = test_wallet.account_delete(account_3)
        self.assertTrue(success)
        self.assertIsNone(test_wallet._default_account)

    def test_wallet_account_delete_by_label(self):
        label_1 = 'Account 1'
        label_2 = 'Account 2'
        label_not_used = 'Account 3'

        password = '123123'
        account_1 = wallet.Account(password=password,
                                   label=label_1)
        account_2 = wallet.Account(password=password,
                                   label=label_2)
        account_3 = wallet.Account(password=password)

        test_wallet = nep6.NEP6DiskWallet.default()
        test_wallet.account_add(account_1)
        test_wallet.account_add(account_2)
        test_wallet.account_add(account_3)

        self.assertEqual(account_1, test_wallet._default_account)

        # delete by label when account is not the default
        success = test_wallet.account_delete_by_label(label_2)
        self.assertTrue(success)
        self.assertEqual(account_1, test_wallet._default_account)

        # delete label not included
        success = test_wallet.account_delete_by_label(label_not_used)
        self.assertFalse(success)

        # delete by label when account is default
        success = test_wallet.account_delete_by_label(label_1)
        self.assertTrue(success)
        self.assertNotEqual(account_1, test_wallet._default_account)
        self.assertEqual(account_3, test_wallet._default_account)

    def test_from_json_with_multisig_account(self):
        p = os.path.join(os.path.dirname(__file__), 'rc2-wallet.json')
        with open(p) as f:
            data = json.load(f)

        w = wallet.Wallet.from_json(data, '123')
        self.assertEqual(2, len(w.accounts))
        self.assertEqual("NY9qiu8YScTM9oAc3nnaeNjaX5fnraaRTA", w.accounts[0].address)
        self.assertEqual("NcmoFiYqThZJFiEYVF1BjYEk6YwF5vtkFA", w.accounts[1].address)



from datetime import datetime


class FlightInfo:
    """
    An internal class for tracking a single outstanding header or block request from a specific node.

    It is used as part of ``SyncManager`` for syncing the chain over the P2P network in combination with the global data
    tracking :class:`RequestInfo <neo3.network.conveniencerequestinfo.RequestInfo>` class.
    """
    def __init__(self, node_id: int, height: int):
        """
        Args:
            node_id: the :attr:`~neo3.network.node.NeoNode.id` of the node the data is requested from
            height: the header or block height being requested
        """

        #: The :attr:`~neo3.network.node.NeoNode.id` of the node the data is requested from.
        #: Defaults to `node_id` parameter.
        self.node_id = node_id

        #: The header or block height being requested.
        #: Defaults to `height` parameter.
        self.height = height

        #: float: UTC timestamp when the instance was created.
        self.start_time: float = datetime.utcnow().timestamp()

    def reset_start_time(self) -> None:
        """ Reset the flight start time."""
        self.start_time = datetime.utcnow().timestamp()



import unittest
import binascii
from neo3.network import capabilities
from neo3.core import serialization

class FullNodeCapabilitiesTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        FullNodeCapability capability = new FullNodeCapability(123);

        Console.WriteLine($"{capability.Size}");
        Console.WriteLine($"{BitConverter.ToString(capability.ToArray()).Replace("-", "")}");
        """
        cls.capability = capabilities.FullNodeCapability(start_height=123)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 5
        self.assertEqual(expected_len, len(self.capability))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'107B000000')
        self.assertEqual(expected_data, self.capability.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_capability = capabilities.FullNodeCapability.deserialize_from_bytes(self.capability.to_array())
        self.assertEqual(self.capability.start_height, deserialized_capability.start_height)

class ServerNodeCapabilitiesTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
            ServerCapability capability = new ServerCapability(NodeCapabilityType.TcpServer, 10333);

            Console.WriteLine($"{capability.Size}");
            Console.WriteLine($"{BitConverter.ToString(capability.ToArray()).Replace("-", "")}");
        """
        cls.capability = capabilities.ServerCapability(n_type=capabilities.NodeCapabilityType.TCPSERVER, port=10333)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 3
        self.assertEqual(expected_len, len(self.capability))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'015D28')
        self.assertEqual(expected_data, self.capability.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_capability = capabilities.ServerCapability.deserialize_from_bytes(self.capability.to_array())
        self.assertEqual(capabilities.NodeCapabilityType.TCPSERVER, deserialized_capability.type)
        self.assertEqual(self.capability.type, deserialized_capability.type)
        self.assertEqual(self.capability.port, deserialized_capability.port)

    def test_creation_with_invalid_type(self):
        with self.assertRaises(TypeError) as context:
            capabilities.ServerCapability(n_type=999, port=123)
        self.assertIn("999 not one of: TCPSERVER WSSERVER", str(context.exception))

class BaseCapabilitiesTestCase(unittest.TestCase):
    def test_deserialize_from(self):
        server = capabilities.ServerCapability(n_type=capabilities.NodeCapabilityType.TCPSERVER, port=10333)
        fullnode = capabilities.FullNodeCapability(start_height=123)

        with serialization.BinaryReader(server.to_array()) as br:
            capability = capabilities.NodeCapability.deserialize_from(br)
        self.assertIsInstance(capability, capabilities.ServerCapability)

        with serialization.BinaryReader(fullnode.to_array()) as br:
            capability = capabilities.NodeCapability.deserialize_from(br)
        self.assertIsInstance(capability, capabilities.FullNodeCapability)

        # test equality
        # ServerCapability capability = new ServerCapability(NodeCapabilityType.TcpServer, 10333);
        # ServerCapability capability2 = new ServerCapability(NodeCapabilityType.TcpServer, 10333);
        # Console.WriteLine($"{capability == capability2} {capability.Equals(capability2)}"); // false false
        fullnode2 = capabilities.FullNodeCapability(start_height=123)
        self.assertNotEqual(fullnode, fullnode2)
        self.assertFalse(fullnode == fullnode2)
        self.assertFalse(id(fullnode) == id(fullnode2))



from __future__ import annotations

import binascii
from typing import List, cast, Tuple
from copy import deepcopy
from neo3 import contracts, vm
from neo3.core import cryptography, to_script_hash, types, serialization, IClonable, IInteroperable, Size as s


class Contract:
    def __init__(self, script: bytes, parameter_list: List[contracts.ContractParameterType]):
        #: The contract instructions (OpCodes)
        self.script = script
        self.parameter_list = parameter_list
        self._script_hash = to_script_hash(self.script)
        self._address = None

    @property
    def script_hash(self) -> types.UInt160:
        """
        The contract script hash
        """
        return self._script_hash

    @classmethod
    def create_multisig_contract(cls, m: int, public_keys: List[cryptography.ECPoint]) -> Contract:
        """
        Create a multi-signature contract requiring `m` signatures from the list `public_keys`.

        Args:
            m: minimum number of signature required for signing. Can't be lower than 2.
            public_keys: public keys to use during verification.
        """
        return cls(script=cls.create_multisig_redeemscript(m, public_keys),
                   parameter_list=[contracts.ContractParameterType.SIGNATURE] * m)

    @staticmethod
    def create_multisig_redeemscript(m: int, public_keys: List[cryptography.ECPoint]) -> bytes:
        """
        Create a multi-signature redeem script requiring `m` signatures from the list `public_keys`.

        This generated script is intended to be executed by the VM to indicate that the requested action is allowed.

        Args:
            m: minimum number of signature required for signing. Can't be lower than 2.
            public_keys: public keys to use during verification.

        Raises:
            ValueError: if the minimum required signatures is not met.
            ValueError: if the maximum allowed signatures is exceeded.
            ValueError: if the maximum allowed public keys is exceeded.
        """
        if m < 1:
            raise ValueError(f"Minimum required signature count is 1, specified {m}.")

        if m > len(public_keys):
            raise ValueError("Invalid public key count. "
                             "Minimum required signatures is bigger than supplied public keys count.")

        if len(public_keys) > 1024:
            raise ValueError(f"Supplied public key count ({len(public_keys)}) exceeds maximum of 1024.")

        sb = vm.ScriptBuilder()
        sb.emit_push(m)
        public_keys.sort()

        for key in public_keys:
            sb.emit_push(key.encode_point(True))

        sb.emit_push(len(public_keys))
        sb.emit_syscall(contracts.syscall_name_to_int("System.Crypto.CheckMultisig"))
        return sb.to_array()

    @classmethod
    def create_signature_contract(cls, public_key: cryptography.ECPoint) -> Contract:
        """
        Create a signature contract.

        Args:
            public_key: the public key to use during verification.

        Returns:

        """
        return cls(cls.create_signature_redeemscript(public_key), [contracts.ContractParameterType.SIGNATURE])

    @staticmethod
    def create_signature_redeemscript(public_key: cryptography.ECPoint) -> bytes:
        """
        Create a single signature redeem script.

        This generated script is intended to be executed by the VM to indicate that the requested action is allowed.

        Args:
            public_key: the public key to use during verification.
        """
        sb = vm.ScriptBuilder()
        sb.emit_push(public_key.encode_point(True))
        sb.emit_syscall(contracts.syscall_name_to_int("System.Crypto.CheckSig"))
        return sb.to_array()

    @staticmethod
    def is_signature_contract(script: bytes) -> bool:
        """
        Test if the provided script is signature contract.

        Args:
            script: contract script.
        """
        if len(script) != 40:
            return False

        if (script[0] != vm.OpCode.PUSHDATA1
                or script[1] != 33
                or script[35] != vm.OpCode.SYSCALL
                or int.from_bytes(script[36:40], 'little') != contracts.syscall_name_to_int(
                    "System.Crypto.CheckSig")):
            return False
        return True

    @staticmethod
    def is_multisig_contract(script: bytes) -> bool:
        """
        Test if the provided script is multi-signature contract.

        Args:
            script: contract script.
        """
        valid, _, _ = Contract.parse_as_multisig_contract(script)
        return valid

    @staticmethod
    def parse_as_multisig_contract(script: bytes) -> Tuple[bool, int, List[cryptography.ECPoint]]:
        """
        Try to parse script as multisig contract and extract related data.

        Args:
            script: array of vm byte code

        Returns:
            bool: True if the script passes as a valid multisignature contract script. False otherwise
            int: the signing threshold if validation passed. 0 otherwise
            list[ECPoint]: the public keys in the script if valiation passed. An empty array otherwise.
        """
        script = bytes(script)
        VALIDATION_FAILURE: Tuple[bool, int, List[cryptography.ECPoint]] = (False, 0, [])

        len_script = len(script)
        if len_script < 42:
            return VALIDATION_FAILURE

        # read signature length, which is encoded as variable_length
        first_byte = script[0]
        if first_byte == int(vm.OpCode.PUSHINT8):
            signature_threshold = script[1]
            i = 2
        elif first_byte == int(vm.OpCode.PUSHINT16):
            signature_threshold = int.from_bytes(script[1:3], 'little', signed=False)
            i = 3
        elif int(vm.OpCode.PUSH1) <= first_byte <= int(vm.OpCode.PUSH16):
            signature_threshold = first_byte - int(vm.OpCode.PUSH0)
            i = 1
        else:
            return VALIDATION_FAILURE

        if signature_threshold < 1 or signature_threshold > 1024:
            return VALIDATION_FAILURE

        # try reading public keys and do a basic format validation
        pushdata1 = int(vm.OpCode.PUSHDATA1)
        public_keys = []
        while script[i] == pushdata1:
            if len_script <= i + 35:
                return VALIDATION_FAILURE
            if script[i + 1] != 33:
                return VALIDATION_FAILURE
            public_keys.append(cryptography.ECPoint.deserialize_from_bytes(script[i + 2:i + 2 + 33]))
            i += 35

        public_key_count = len(public_keys)
        if public_key_count < signature_threshold or public_key_count > 1024:
            return VALIDATION_FAILURE

        # validate that the number of collected public keys match the expected count
        value = script[i]
        if value == int(vm.OpCode.PUSHINT8):
            if len_script <= i + 1 or public_key_count != script[i + 1]:
                return VALIDATION_FAILURE
            i += 2
        elif value == int(vm.OpCode.PUSHINT16):
            if len_script < i + 3 or public_key_count != int.from_bytes(script[i + 1:i + 3], 'little', signed=False):
                return VALIDATION_FAILURE
            i += 3
        elif int(vm.OpCode.PUSH1) <= value <= int(vm.OpCode.PUSH16):
            if public_key_count != value - int(vm.OpCode.PUSH0):
                return VALIDATION_FAILURE
            i += 1
        else:
            return VALIDATION_FAILURE

        if len_script != i + 5:
            return VALIDATION_FAILURE

        if script[i] != int(vm.OpCode.SYSCALL):
            return VALIDATION_FAILURE
        i += 1

        syscall_num = int.from_bytes(script[i:i + 4], 'little')
        if syscall_num != contracts.syscall_name_to_int("System.Crypto.CheckMultisig"):
            return VALIDATION_FAILURE
        return True, signature_threshold, public_keys

    @staticmethod
    def get_consensus_address(validators: List[cryptography.ECPoint]) -> types.UInt160:
        script = contracts.Contract.create_multisig_redeemscript(
            len(validators) - (len(validators) - 1) // 3,
            validators
        )
        return to_script_hash(script)


class ContractState(serialization.ISerializable, IClonable, IInteroperable):
    def __init__(self,
                 id_: int,
                 nef: contracts.NEF,
                 manifest_: contracts.ContractManifest,
                 update_counter: int,
                 hash_: types.UInt160):
        self.id = id_
        self.nef = nef
        self.manifest = manifest_
        self.update_counter = update_counter
        self.hash = hash_

    def __len__(self):
        return (s.uint32  # id
                + len(self.nef.to_array())
                + len(self.manifest)
                + s.uint16  # update counter
                + len(self.hash))

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash != other.hash:
            return False
        return True

    def __deepcopy__(self, memodict={}):
        return ContractState.deserialize_from_bytes(self.to_array())

    @property
    def script(self) -> bytes:
        return self.nef.script

    @script.setter
    def script(self, value: bytes) -> None:
        self.nef.script = value

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_int32(self.id)
        writer.write_serializable(self.nef)
        writer.write_serializable(self.manifest)
        writer.write_uint16(self.update_counter)
        writer.write_serializable(self.hash)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.id = reader.read_int32()
        self.nef = reader.read_serializable(contracts.NEF)
        self.manifest = reader.read_serializable(contracts.ContractManifest)
        self.update_counter = reader.read_uint16()
        self.hash = reader.read_serializable(types.UInt160)

    def from_replica(self, replica):
        super().from_replica(replica)
        self.id = replica.id
        self.nef = replica.nef
        self.manifest = replica.manifest
        self.update_counter = replica.update_counter
        self.hash = replica.hash

    def clone(self):
        return ContractState(self.id, deepcopy(self.nef), deepcopy(self.manifest), self.update_counter, self.hash)

    @classmethod
    def from_stack_item(cls, stack_item: vm.StackItem):
        array = cast(vm.ArrayStackItem, stack_item)
        id = int(array[0].to_biginteger())
        update_counter = int(array[1].to_biginteger())
        hash_ = types.UInt160(array[2].to_array())
        nef = contracts.NEF.deserialize_from_bytes(array[3].to_array())
        manifest = contracts.ContractManifest.deserialize_from_bytes(array[4].to_array())
        return cls(id, nef, manifest, update_counter, hash_)

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        array = vm.ArrayStackItem(reference_counter)
        id_ = vm.IntegerStackItem(self.id)
        nef = vm.ByteStringStackItem(self.nef.to_array())
        update_counter = vm.IntegerStackItem(self.update_counter)
        hash_ = vm.ByteStringStackItem(self.hash.to_array())
        array.append([id_, update_counter, hash_, nef, self.manifest.to_stack_item(reference_counter)])
        return array

    def can_call(self, target_contract: ContractState, target_method: str) -> bool:
        results = list(map(lambda p: p.is_allowed(target_contract, target_method), self.manifest.permissions))
        return any(results)

    @classmethod
    def _serializable_init(cls):
        return cls(0, contracts.NEF._serializable_init(), contracts.ContractManifest(), 0, types.UInt160.zero())



from __future__ import annotations
from typing import Optional
from . import register, NativeContract
from neo3 import storage, contracts, vm
from neo3.core import types
from neo3.network import payloads


class LedgerContract(NativeContract):
    _id = -4

    def init(self):
        super(LedgerContract, self).init()

    @register("currentHash", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def current_hash(self, snapshot: storage.Snapshot) -> types.UInt256:
        """ Get the hash of the current block """
        return snapshot.persisting_block.hash()

    @register("currentIndex", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def current_index(self, snapshot: storage.Snapshot) -> int:
        """ Get the block height of the current block """
        return snapshot.best_block_height

    @register("getBlock", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_block(self, snapshot: storage.Snapshot, index_or_hash: bytes) -> Optional[payloads.TrimmedBlock]:
        """
        Fetch a block from storage.

        Args:
            snapshot: the snapshot to grab the data from.
            index_or_hash: the height or block hash of the block we wish to retrieve

        Raises:
             ValueError: if the height is invalid (negative or too large)
             ValueError: if the index_or_hash field could not be converted to a valid height or hash.
        """
        if len(index_or_hash) < types.UInt256._BYTE_LEN:
            height = vm.BigInteger(index_or_hash)
            if height < 0 or height > 4294967295:  # uint.MaxValue
                raise ValueError("Invalid height")
            block = snapshot.blocks.try_get_by_height(int(height), read_only=True)
        elif len(index_or_hash) == types.UInt256._BYTE_LEN:
            block_hash = types.UInt256(index_or_hash)
            block = snapshot.blocks.try_get(block_hash, read_only=True)
        else:
            raise ValueError("Invalid data")

        if block is None or not self._is_traceable_block(snapshot, block.index):
            return None
        return block.trim()

    @register("getTransaction", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_tx_for_contract(self, snapshot: storage.Snapshot, hash_: types.UInt256) -> Optional[payloads.Transaction]:
        """ Fetch a transaction from storage by its hash """
        tx = snapshot.transactions.try_get(hash_, read_only=True)
        if tx is None or not self._is_traceable_block(snapshot, tx.block_height):
            return None
        return tx

    @register("getTransactionheight", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_tx_height(self, snapshot: storage.Snapshot, hash_: types.UInt256) -> int:
        """
        Get the height of the block that the transaction is included in.

        Returns:
            -1 if the transaction could not be found. The height otherwise.
        """
        tx = snapshot.transactions.try_get(hash_, read_only=True)
        if tx is None or not self._is_traceable_block(snapshot, tx.block_height):
            return -1
        return tx.block_height

    @register("getTransactionFromBlock", contracts.CallFlags.READ_STATES, cpu_price=1 << 16)
    def get_tx_from_block(self,
                          snapshot: storage.Snapshot,
                          block_index_or_hash: bytes,
                          tx_index: int) -> Optional[payloads.Transaction]:
        """
        Get a transaction from a specific block
        Args:
            snapshot:
            block_index_or_hash: the height or block hash of the block we wish to retrieve
            tx_index: the index into the blocks transaction list of the transaction we want to fetch.

        Returns:

        """
        if len(block_index_or_hash) < types.UInt256._BYTE_LEN:
            height = vm.BigInteger(block_index_or_hash)
            if height < 0 or height > 4294967295:  # uint.MaxValue
                raise ValueError("Invalid height")
            block = snapshot.blocks.try_get_by_height(height, read_only=True)
        elif len(block_index_or_hash) == types.UInt256._BYTE_LEN:
            block_hash = types.UInt256(block_index_or_hash)
            block = snapshot.blocks.try_get(block_hash, read_only=True)
        else:
            raise ValueError("Invalid data")

        if block and not self._is_traceable_block(snapshot, block.index):
            block = None
        if tx_index < 0 or tx_index >= len(block.transactions):
            raise ValueError("Transaction index out of range")
        return block.transactions[tx_index]

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        # Unlike C# the current block or its transactions are not persisted here,
        # it is still done in the Blockchain class in the persist() function
        pass

    def _is_traceable_block(self, snapshot: storage.Snapshot, index: int) -> bool:
        current_idx = self.current_index(snapshot)
        if index > current_idx:
            return False
        # otherwise limit search back distance
        MAX_TRACABLE_BLOCKS = 2_102_400
        return index + MAX_TRACABLE_BLOCKS > current_idx



from __future__ import annotations
import hashlib
from neo3 import contracts
from typing import Dict, Callable, Optional, get_type_hints


class InteropDescriptor:
    NONE_TYPE = type(None)

    def __init__(self,
                 method: str,
                 handler: Callable,
                 price: int,
                 call_flags: contracts.CallFlags):
        """
        Create a interoperability call descriptor.
        This are the functions that can be called using the SYSCALL OpCode in the virtual machine.

        Use the alternative constructor `create_with_price_calculator` if the price needs to be determined dynamically.

        Args:
            method: name of call.
            handler: the function that will be executed when called.
            price: the price of calling the handler.
            call_flags: ExecutionContext rights needed.
        """
        self.method = method
        self.hash: int = int.from_bytes(hashlib.sha256(self.method.encode()).digest()[:4], 'little', signed=False)
        self.handler = handler
        self.parameters = []
        for k, v in get_type_hints(handler).items():
            if k == 'return':
                self.has_return_value = v != self.NONE_TYPE
                continue
            self.parameters.append(v)

        # while using the @register decorator, the first argument to the function is always the application engine
        # itself, we want to strip that off
        self.parameters = self.parameters[1:]
        self.price = price
        self.required_call_flags = call_flags


class InteropService:
    _methods: Dict[int, InteropDescriptor] = {}

    @classmethod
    def get_descriptor(cls, method_id) -> Optional[InteropDescriptor]:
        return cls._methods.get(method_id, None)

    @classmethod
    def register(cls,
                 method: str,
                 handler: Callable,
                 fixed_price: int,
                 call_flags: contracts.CallFlags) -> InteropDescriptor:
        """
        Register an interoperability method to the interoperability service that can be called by the Virtual Machine.

        Args:
            method: name of call.
            handler: the function that will be executed when called.
            fixed_price: the price for calling the handler.
            call_flags: ExecutionContext rights needed.
        """
        descriptor = InteropDescriptor(method, handler, fixed_price, call_flags)
        cls._methods.update({descriptor.hash: descriptor})
        return descriptor



from __future__ import annotations
from neo3.network import payloads
from neo3.core import utils as core_utils
from neo3 import contracts, storage, vm, wallet
from neo3.contracts.interop.crypto import CHECKSIG_PRICE


def add_system_fee(tx: payloads.Transaction, snapshot: storage.Snapshot) -> None:
    tx.system_fee = calculate_system_fee(tx, snapshot)


def add_network_fee(tx: payloads.Transaction, snapshot: storage.Snapshot, account: wallet.Account) -> None:
    tx.network_fee = calculate_network_fee(tx, snapshot, account)


def calculate_system_fee(tx: payloads.Transaction, snapshot: storage.Snapshot) -> int:
    engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, tx, snapshot, 0, test_mode=True)
    engine.load_script(vm.Script(tx.script))
    if engine.execute() == vm.VMState.FAULT:
        raise ValueError("Transaction script execution failed")
    else:
        return engine.gas_consumed


def calculate_network_fee(tx: payloads.Transaction, snapshot: storage.Snapshot, account: wallet.Account) -> int:
    if len(tx.signers) == 0:
        raise ValueError("Cannot calculate the network fee without a sender in the transaction.")

    hashes = tx.get_script_hashes_for_verifying(snapshot)
    network_fee_size = (tx.HEADER_SIZE
                        + core_utils.get_var_size(tx.signers)  # type: ignore
                        + core_utils.get_var_size(tx.attributes)  # type: ignore
                        + core_utils.get_var_size(tx.script)  # type: ignore
                        + core_utils.get_var_size(len(hashes))  # type: ignore
                        )
    exec_fee_factor = contracts.PolicyContract().get_exec_fee_factor(snapshot)

    network_fee = 0
    for i, hash_ in enumerate(hashes):
        witness_script = None
        if hash_ == account.script_hash and account.contract and len(account.contract.script) > 0:
            witness_script = account.contract.script

        if witness_script is None and len(tx.witnesses) > 0:
            for witness in tx.witnesses:
                if witness.script_hash() == hash_:
                    witness_script = witness.verification_script
                    break

        if witness_script is None or (witness_script and len(witness_script) == 0):
            raise ValueError("Using a smart contract as a witness is not yet supported in mamba")

        elif contracts.Contract.is_signature_contract(witness_script):
            network_fee_size += 67 + core_utils.get_var_size(witness_script)  # type: ignore
            network_fee = exec_fee_factor * signature_contract_costs()
        elif contracts.Contract.is_multisig_contract(witness_script):
            _, threshold, public_keys = contracts.Contract.parse_as_multisig_contract(witness_script)
            invocation_script_size = 66 * threshold
            network_fee_size += (core_utils.get_var_size(invocation_script_size)  # type: ignore
                                 + invocation_script_size
                                 + core_utils.get_var_size(witness_script))  # type: ignore
            network_fee = exec_fee_factor * multisig_contract_costs(threshold, len(public_keys))

    network_fee += network_fee_size * contracts.PolicyContract().get_fee_per_byte(snapshot)
    return network_fee


def signature_contract_costs() -> int:
    return (contracts.ApplicationEngine.opcode_price(vm.OpCode.PUSHDATA1) * 2
            + contracts.ApplicationEngine.opcode_price(vm.OpCode.SYSCALL)
            + CHECKSIG_PRICE
            )


def multisig_contract_costs(threshold: int, public_key_count: int) -> int:
    fee = contracts.ApplicationEngine.opcode_price(vm.OpCode.PUSHDATA1) * (threshold + public_key_count)
    opcode = vm.OpCode(vm.ScriptBuilder().emit_push(threshold).to_array()[0])
    fee += contracts.ApplicationEngine.opcode_price(opcode)
    opcode = vm.OpCode(vm.ScriptBuilder().emit_push(public_key_count).to_array()[0])
    fee += contracts.ApplicationEngine.opcode_price(opcode)
    fee += contracts.ApplicationEngine.opcode_price(vm.OpCode.SYSCALL)
    fee += CHECKSIG_PRICE * public_key_count
    return fee



from __future__ import annotations
import base64
import base58  # type: ignore
from . import NativeContract, register
from neo3 import contracts, vm
from typing import List
import orjson as json  # type: ignore


class StdLibContract(NativeContract):
    _MAX_INPUT_LENGTH = 1024

    _service_name = "StdLib"
    _id = -2

    def init(self):
        super(StdLibContract, self).init()

    @register("serialize", contracts.CallFlags.NONE, cpu_price=1 << 12)
    def binary_serialize(self, engine: contracts.ApplicationEngine, stack_item: vm.StackItem) -> bytes:
        return contracts.BinarySerializer.serialize(stack_item, engine.MAX_ITEM_SIZE)

    @register("deserialize", contracts.CallFlags.NONE, cpu_price=1 << 14)
    def binary_deserialize(self, engine: contracts.ApplicationEngine, data: bytes) -> vm.StackItem:
        return contracts.BinarySerializer.deserialize(data, engine.MAX_ITEM_SIZE, engine.reference_counter)

    @register("jsonSerialize", contracts.CallFlags.NONE, cpu_price=1 << 12)
    def json_serialize(self, engine: contracts.ApplicationEngine, stack_item: vm.StackItem) -> bytes:
        v = bytes(contracts.JSONSerializer.serialize(stack_item, engine.MAX_ITEM_SIZE), 'utf-8')
        v = v.replace(b'+', b'\x5c\x75\x30\x30\x32\x42')
        return v

    @register("jsonDeserialize", contracts.CallFlags.NONE, cpu_price=1 << 14)
    def json_deserialize(self, engine: contracts.ApplicationEngine, data: bytes) -> vm.StackItem:
        return contracts.JSONSerializer.deserialize(json.loads(data.decode()), engine.reference_counter)

    @register("itoa", contracts.CallFlags.NONE, cpu_price=1 << 12)
    def do_itoa_base10(self, value: vm.BigInteger) -> str:
        return self.do_itoa(value, 10)

    @register("itoa", contracts.CallFlags.NONE, cpu_price=1 << 12)
    def do_itoa(self, value: vm.BigInteger, base: int) -> str:
        if base == 10:
            return str(value)
        elif base == 16:
            return hex(int(value))[2:]
        else:
            raise ValueError("Invalid base specified")

    @register("atoi", contracts.CallFlags.NONE, cpu_price=1 << 6)
    def do_atoi_base10(self, value: str) -> int:
        return self.do_atoi(value, 10)

    @register("atoi", contracts.CallFlags.NONE, cpu_price=1 << 6)
    def do_atoi(self, value: str, base: int) -> int:
        if len(value) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        if base != 10 and base != 16:
            raise ValueError("Invalid base specified")
        else:
            return int(value, base)

    @register("base64Encode", contracts.CallFlags.NONE, cpu_price=1 << 5)
    def base64_encode(self, data: bytes) -> str:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base64.b64encode(data).decode()

    @register("base64Decode", contracts.CallFlags.NONE, cpu_price=1 << 5)
    def base64_decode(self, data: bytes) -> bytes:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base64.b64decode(data)

    @register("base58Encode", contracts.CallFlags.NONE, cpu_price=1 << 13)
    def base58_encode(self, data: bytes) -> str:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base58.b58encode(data).decode()

    @register("base58Decode", contracts.CallFlags.NONE, cpu_price=1 << 10)
    def base58_decode(self, data: bytes) -> bytes:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base58.b58decode(data)

    @register("base58CheckEncode", contracts.CallFlags.NONE, cpu_price=1 << 16)
    def base58_check_encode(self, data: bytes) -> str:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base58.b58encode_check(data).decode()

    @register("base58CheckDecode", contracts.CallFlags.NONE, cpu_price=1 << 16)
    def base58_check_decode(self, data: bytes) -> bytes:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        return base58.b58decode_check(data)

    @register("memoryCompare", contracts.CallFlags.NONE, cpu_price=1 << 5)
    def memory_compare(self, str1: bytes, str2: bytes) -> int:
        if len(str1) > self._MAX_INPUT_LENGTH or len(str2) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"One or both arguments exceed maximum length of {self._MAX_INPUT_LENGTH}")
        if str1 < str2:
            return -1
        elif str1 == str2:
            return 0
        else:
            return 1

    @register("memorySearch", contracts.CallFlags.NONE, cpu_price=1 << 6)
    def memory_search1(self, memory: bytes, needle: bytes) -> int:
        return self.memory_search(memory, needle, 0, False)

    @register("memorySearch", contracts.CallFlags.NONE, cpu_price=1 << 6)
    def memory_search2(self, memory: bytes, needle: bytes, start: int) -> int:
        return self.memory_search(memory, needle, start, False)

    @register("memorySearch", contracts.CallFlags.NONE, cpu_price=1 << 6)
    def memory_search(self, memory: bytes, needle: bytes, start: int, backwards: bool) -> int:
        if len(memory) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Value exceeds maximum length of {self._MAX_INPUT_LENGTH}")

        if len(memory) == 0:
            return -1

        if backwards:
            return memory[:start].rfind(needle)
        else:
            idx = memory[start:].find(needle)
            return idx + start

    @register("stringSplit", contracts.CallFlags.NONE, cpu_price=1 << 8)
    def string_split1(self, data: str, separator: str) -> List[str]:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Input value exceeds maximum length of {self._MAX_INPUT_LENGTH}")
        if len(separator) == 0 or separator is None:
            raise ValueError(f"Invalid separator in std.stringSplit")
        return data.split(separator)

    @register("stringSplit", contracts.CallFlags.NONE, cpu_price=1 << 8)
    def string_split2(self, data: str, separator: str, remove_empty_entries: bool) -> List[str]:
        if len(data) > self._MAX_INPUT_LENGTH:
            raise ValueError(f"Input value exceeds maximum length of {self._MAX_INPUT_LENGTH}")

        if len(separator) == 0 or separator is None:
            raise ValueError(f"Invalid separator in std.stringSplit")
        result = data.split(separator)
        if remove_empty_entries:
            result = [x for x in result if x != '']
        return result



import unittest
from neo3 import vm, contracts


class BinarySerializerTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    @classmethod
    def setUpClass(cls) -> None:
        cls.reference_counter = vm.ReferenceCounter()

    def test_with_map(self):
        """
        var m = new Map(new ReferenceCounter());
        var i = new Integer(new BigInteger(1));
        var i2 = new Integer(new BigInteger(2));
        var b = new Neo.VM.Types.Boolean(true);
        m[i] = b;
        m[i2] = b;

        """
        # moved outside of multiline comment because pycharm is broken: https://youtrack.jetbrains.com/issue/PY-43117
        #Console.WriteLine($"b'\\x{BitConverter.ToString(BinarySerializer.Serialize(m, 999)).Replace("-", @"\x")}'");
        m = vm.MapStackItem(self.reference_counter)
        i = vm.IntegerStackItem(vm.BigInteger(1))
        i2 = vm.IntegerStackItem(vm.BigInteger(2))
        b = vm.BooleanStackItem(True)
        m[i] = b
        m[i2] = b
        expected = b'\x48\x02\x21\x01\x01\x20\x01\x21\x01\x02\x20\x01'
        out = contracts.BinarySerializer.serialize(m, 999)
        self.assertEqual(expected, out)

        # now we add a reference to ourselves.
        with self.assertRaises(ValueError) as context:
            m[i] = m
            contracts.BinarySerializer.serialize(m, 999)
        self.assertEqual("Item already exists", str(context.exception))

        # now test deserialization
        m[i] = b # restore m[i] to original content
        new_m = contracts.BinarySerializer.deserialize(out, 2048, self.reference_counter)
        self.assertEqual(len(m), len(new_m))
        self.assertEqual(m.keys(), new_m.keys())
        self.assertEqual(m.values(), new_m.values())

    def test_with_array(self):
        """
        var a = new Neo.VM.Types.Array();
        var i = new Integer(new BigInteger(1));
        var i2 = new Integer(new BigInteger(2));
        a.Add(i);
        a.Add(i2);
        """
        # Console.WriteLine($"b'\\x{BitConverter.ToString(BinarySerializer.Serialize(a, 999)).Replace("-", @"\x")}'");
        # moved outside of multiline comment because pycharm is broken: https://youtrack.jetbrains.com/issue/PY-43117
        i = vm.IntegerStackItem(vm.BigInteger(1))
        i2 = vm.IntegerStackItem(vm.BigInteger(2))
        a = vm.ArrayStackItem(self.reference_counter)
        a.append(i)
        a.append(i2)
        expected = b'\x40\x02\x21\x01\x01\x21\x01\x02'
        out = contracts.BinarySerializer.serialize(a, 999)
        self.assertEqual(expected, out)

        # now we add a reference to ourselves.
        with self.assertRaises(ValueError) as context:
            a.append(a)
            contracts.BinarySerializer.serialize(a, 999)
        self.assertEqual("Item already exists", str(context.exception))

        # now test deserialization
        # first remove the reference to self
        a.remove(len(a) - 1)
        new_a = contracts.BinarySerializer.deserialize(out, 2048, self.reference_counter)
        self.assertIsInstance(a, vm.ArrayStackItem)
        self.assertEqual(a._items, new_a._items)

    def test_with_null(self):
        n = vm.NullStackItem()
        x = contracts.BinarySerializer.serialize(n, 999)
        self.assertEqual(b'\x00', x)

        new_n = contracts.BinarySerializer.deserialize(x, 1, self.reference_counter)
        self.assertIsInstance(new_n, vm.NullStackItem)

    def test_serialize_with_invalid_stackitem(self):
        s = vm.Script(b'')
        p = vm.PointerStackItem(s, 0)

        with self.assertRaises(ValueError) as context:
            contracts.BinarySerializer.serialize(p, 999)
        self.assertIn("Cannot serialize", str(context.exception))

    def test_serialize_with_exceeding_max_length(self):
        with self.assertRaises(ValueError) as context:
            n = vm.NullStackItem()
            contracts.BinarySerializer.serialize(n, 0)
        self.assertEqual("Output length exceeds max size", str(context.exception))

    def test_deserialize_invalid_data(self):
        with self.assertRaises(ValueError) as context:
            contracts.BinarySerializer.deserialize(b'', 1, self.reference_counter)
        self.assertEqual("Nothing to deserialize", str(context.exception))

    def test_deserialize_bytestring(self):
        data = b'\x01\x02'
        b = vm.ByteStringStackItem(data)
        b_serialized = contracts.BinarySerializer.serialize(b, 999)
        new_b = contracts.BinarySerializer.deserialize(b_serialized, 999, self.reference_counter)
        self.assertIsInstance(new_b, vm.ByteStringStackItem)
        self.assertEqual(new_b, b)

    def test_deserialize_buffer(self):
        data = b'\x01\x02'
        b = vm.BufferStackItem(data)
        b_serialized = contracts.BinarySerializer.serialize(b, 999)
        new_b = contracts.BinarySerializer.deserialize(b_serialized, 999, self.reference_counter)
        self.assertIsInstance(new_b, vm.BufferStackItem)
        self.assertEqual(new_b.to_array(), b.to_array())

    def test_deserialize_struct(self):
        s = vm.StructStackItem(self.reference_counter)
        bool1 = vm.BooleanStackItem(True)
        bool2 = vm.BooleanStackItem(False)
        s.append([bool1, bool2])
        s_serialized = contracts.BinarySerializer.serialize(s, 999)
        new_s = contracts.BinarySerializer.deserialize(s_serialized, 999, self.reference_counter)
        self.assertIsInstance(new_s, vm.StructStackItem)
        for l, r in zip(new_s._items, s._items):
            self.assertEqual(l, r)

    def test_deserialize_with_invalid_format(self):
        data = b'\x01\x02'
        b = vm.BufferStackItem(data)
        b_serialized = bytearray(contracts.BinarySerializer.serialize(b, 999))
        b_serialized[0] = 0xFF # non existing stackitem type
        with self.assertRaises(ValueError) as context:
            contracts.BinarySerializer.deserialize(b_serialized, 999, self.reference_counter)
        self.assertEqual("Invalid format", str(context.exception))


from bitarray import bitarray  # type: ignore
import mmh3  # type: ignore


class BloomFilter:
    """
    BloomFilter implementation conforming to NEO's `implementation <https://github.com/neo-project/neo/blob/982e69090f27c1415872536ce39aea22f0873467/neo.UnitTests/Cryptography/UT_BloomFilter.cs>`_.  # noqa
    """
    def __init__(self, m: int, k: int, ntweak: int, elements: bytes = None):
        """

        Args:
            m: size of bitarray.
            k: number of hash functions.
            ntweak: correction factor.
            elements: hex-escaped bytearray of values to create the bitarray from.
                Warning: the bit array is truncated to size `m`.

        """
        self.K = k
        self.seeds = [(p * 0xFBA4C795 + ntweak) % 4294967296 for p in range(0, k)]
        if elements:
            tmp_bits = bitarray(endian='little')
            tmp_bits.frombytes(elements)
            # truncate to m bits
            self.bits = tmp_bits[:m]
        else:
            self.bits = bitarray(m, endian='little')
            self.bits.setall(False)
        self.tweak = ntweak

    def add(self, element: bytes) -> None:
        """
        Add an element to the filter.

        Args:
            element: hex-escaped bytearray.
        """
        for s in self.seeds:
            h = mmh3.hash(element, s, signed=False)
            self.bits[h % len(self.bits)] = True

    def check(self, element: bytes) -> bool:
        """
        Check if the element is present

        Args:
            element: hex-escaped bytearray

        Returns: True if present. False if not present.
        """
        for s in self.seeds:
            h = mmh3.hash(element, s, signed=False)
            if not self.bits[h % len(self.bits)]:
                return False
        return True

    def get_bits(self) -> bytes:
        """
        Return the filter bits.
        """
        return self.bits.tobytes()
